from ws4py.client.threadedclient import WebSocketClient
import json
import threading
import logging
from .errors import ConnectionClosedError, UnknownRequestError

log = logging.getLogger('macumba')


class JujuWS(WebSocketClient):

    def __init__(self, url, password, protocols=['https-only'],
                 extensions=None, ssl_options=None, headers=None,
                 start_reqid=1):
        WebSocketClient.__init__(self, url, protocols, extensions,
                                 ssl_options=ssl_options, headers=headers)
        self.open_done = threading.Event()
        self.rid_lock = threading.RLock()
        self.msglock = threading.RLock()
        self.messages = {}
        self._cur_request_id = start_reqid

    # WebSocketClient subclass overrides, run in private thread:
    def opened(self):
        self.open_done.set()

    def received_message(self, m):
        msg = json.loads(m.data.decode('utf-8'))
        msg_req_id = msg['RequestId']
        with self.msglock:
            self.messages[msg_req_id] = msg

    def closed(self, code, reason=None):
        log.debug("socket closed: code:{} reason:{}".format(code, reason))

    # actions for users of the class:
    def get_current_request_id(self):
        "only intended to pass to constructor of a replacing client"
        return self._cur_request_id

    def do_close(self):
        self.close()

    def do_connect(self, creds):
        self.connect()
        self.open_done.wait()
        self.open_done.clear()
        rv = self.do_send(creds)
        return rv

    def do_send(self, json_message):
        with self.rid_lock:
            self._cur_request_id += 1
            request_id = self._cur_request_id

        json_message['RequestId'] = request_id

        self.send(json.dumps(json_message))

        with self.msglock:
            self.messages[request_id] = None

        return request_id

    def do_receive(self, request_id):
        """Checks for message matching request_id.

        Will return None if message has not arrived yet.

        Raises UnknownRequestError if request_id hasn't been sent yet
        (or was already received).

        """
        if self.terminated:
            raise ConnectionClosedError

        with self.msglock:
            if request_id not in self.messages:
                errmsg = ("{} not in messages. "
                          "cur = {}".format(request_id,
                                            self._cur_request_id))
                raise UnknownRequestError(errmsg)

            message = self.messages[request_id]
            if message is not None:
                del self.messages[request_id]

        return message



# Copyright 2016 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from urwid import AttrMap, Divider, Pile, Text, WidgetWrap

from ubuntui.widgets.buttons import MenuSelectButton

from bundleplacer.relationtype import RelationType

import logging

log = logging.getLogger('bundleplacer')


class RelationWidget(WidgetWrap):

    def __init__(self, source_service_name, source_relname, interface,
                 reltype, target_service_name, target_relname,
                 placement_controller, select_cb):
        self.source_service_name = source_service_name
        self.source_relname = source_relname
        self.interface = interface
        self.reltype = reltype
        self.target_service_name = target_service_name
        self.target_relname = target_relname
        self.select_cb = select_cb
        self.placement_controller = placement_controller
        w = self.build_widgets()
        super().__init__(w)
        self.update()

    def selectable(self):
        return True

    def build_widgets(self):

        self.button = MenuSelectButton("",
                                       on_press=self.do_select)
        return AttrMap(self.button, 'text',
                       'button_secondary focus')

    def update(self):
        arrow = {RelationType.Provides: "\N{RIGHTWARDS ARROW}",
                 RelationType.Requires:
                 "\N{LEFTWARDS ARROW}"}[self.reltype]

        pc = self.placement_controller
        self.connected = pc.is_related(self.source_service_name,
                                       self.source_relname,
                                       self.target_service_name,
                                       self.target_relname)

        connstr = {True: "\N{CHECK MARK} ",
                   False: "  "}[self.connected]

        title = "{} {} {}  {}:{}".format(connstr,
                                         self.source_relname,
                                         arrow,
                                         self.target_service_name,
                                         self.target_relname)
        self.button.set_label(title)

    def do_select(self, sender):
        self.select_cb(self.source_relname, self.target_service_name,
                       self.target_relname)


class NoRelationWidget(WidgetWrap):
    def __init__(self, relname, iface, reltype):
        other_reltype = {RelationType.Requires: RelationType.Provides,
                         RelationType.Provides: RelationType.Requires}[reltype]
        self.source_relname = relname
        self.reltype = reltype
        s = "({}: nothing {} {})".format(
            relname, other_reltype.name.lower(), iface)
        super().__init__(AttrMap(MenuSelectButton(s), 'label',
                                 'button_secondary focus'))

    def selectable(self):
        return True

    def update(self):
        "no op"


class RelationsColumn(WidgetWrap):

    """UI to edit relations of a service
    """

    def __init__(self, display_controller, placement_controller,
                 placement_view, metadata_controller):
        self.placement_controller = placement_controller
        self.metadata_controller = metadata_controller
        self.service = None
        self.provides = set()
        self.requires = set()
        self.placement_view = placement_view
        w = self.build_widgets()
        super().__init__(w)
        self.update()

    def build_widgets(self):
        self.title = Text('')
        self.relation_widgets = []
        self.pile = Pile([Divider(), self.title] + self.relation_widgets)
        return self.pile

    def refresh(self):
        self.set_service(self.service)

    def add_charm(self, charm_name):
        self.metadata_controller.add_charm(charm_name)

    def set_service(self, service):
        self.service = service
        self.add_charm(service.charm_name)
        self.pile.contents = self.pile.contents[:2]
        self.provides = set()
        self.requires = set()
        self.relation_widgets = []

    def update(self):
        if self.service is None:
            return

        self.title.set_text(('body', "Edit relations for {}".format(
            self.service.service_name)))

        if len(self.relation_widgets) == 0:
            self.title.set_text(('body', "Loading Relations..."))
        else:
            self.title.set_text(('body', "Select Relations:"))

        p = set(self.metadata_controller.get_provides(self.service.charm_name))
        r = set(self.metadata_controller.get_requires(self.service.charm_name))
        new_provides = p - self.provides
        self.provides.update(p)
        new_requires = r - self.requires
        self.requires.update(r)

        mc = self.metadata_controller
        args = [(relname, iface, RelationType.Provides,
                 mc.get_services_for_iface(iface, RelationType.Requires))
                for relname, iface in sorted(new_provides)]

        args += [(relname, iface, RelationType.Requires,
                  mc.get_services_for_iface(iface, RelationType.Provides))
                 for relname, iface in sorted(new_requires)]

        for relname, iface, reltype, matches in args:
            if len(matches) == 0:
                rw = NoRelationWidget(relname, iface, reltype)
                self.relation_widgets.append(rw)
                self.pile.contents.append((rw,
                                           self.pile.options()))

            for tgt_relname, tgt_service in matches:
                if tgt_service.service_name == self.service.service_name:
                    continue
                rw = RelationWidget(self.service.service_name, relname,
                                    iface, reltype, tgt_service.service_name,
                                    tgt_relname,
                                    self.placement_controller,
                                    self.do_select)
                self.relation_widgets.append(rw)
                self.pile.contents.append((rw,
                                           self.pile.options()))

        for w, _ in self.pile.contents[2:]:
            w.update()
        self.sort_relation_widgets()

    def focus_prev_or_top(self):
        # ? self.pile.focus_position = len(self.pile.contents) - 1

        if len(self.pile.contents) <= 2:
            return
        pos = self.pile.focus_position
        if pos < 2:
            self.pile.focus_position = 2

    def do_select(self, source_relname, tgt_service_name,
                  tgt_relation_name):
        self.placement_controller.toggle_relation(self.service.service_name,
                                                  source_relname,
                                                  tgt_service_name,
                                                  tgt_relation_name)
        self.update()

    def sort_relation_widgets(self):
        def keyfunc(rw):
            return str(rw.reltype) + rw.source_relname
        self.relation_widgets.sort(key=keyfunc)

        def wrappedkeyfunc(t):
            rw, options = t
            if isinstance(rw, RelationWidget):
                return keyfunc(rw)
            if isinstance(rw, NoRelationWidget):
                return 'z' + keyfunc(rw)
            return 'A'

        self.pile.contents.sort(key=wrappedkeyfunc)



# Copyright (c) 2013, Daniel Lindsley.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#
#     3. Neither the name of shell nor the names of its contributors may be used
#        to endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
shell
=====

A better way to run shell commands in Python.

If you just need to quickly run a command, you can use the ``shell`` shortcut
function::

    >>> from shell import shell
    >>> ls = shell('ls')
    >>> for file in ls.output():
    ...     print file
    'another.txt'

If you need to extend the behavior, you can also use the ``Shell`` object::

    >>> from shell import Shell
    >>> sh = Shell(has_input=True)
    >>> cat = sh.run('cat -u')
    >>> cat.write('Hello, world!')
    >>> cat.output()
    ['Hello, world!']

"""
import shlex
import subprocess


__author__ = 'Daniel Lindsley'
__license__ = 'New BSD'
__version__ = (1, 0, 1)


class ShellException(Exception):
    """The base exception for all shell-related errors."""
    pass


class MissingCommandException(ShellException):
    """Thrown when no command was setup."""
    pass


class CommandError(ShellException):
    """Thrown when a command fails."""
    error_code = 1


class Shell(object):
    """
    Handles executing commands & recording output.

    Optionally accepts a ``has_input`` parameter, which should be a boolean.
    If set to ``True``, the command will wait to execute until you call the
    ``Shell.write`` method & send input. (Default: ``False``)

    Optionally accepts a ``record_output`` parameter, which should be a
    boolean.
    If set to ``True``, the stdout from the command will be recorded.
    (Default: ``True``)

    Optionally accepts a ``record_errors`` parameter, which should be a
    boolean.
    If set to ``True``, the stderr from the command will be recorded.
    (Default: ``True``)

    Optionally accepts a ``strip_empty`` parameter, which should be a boolean.
    If set to ``True``, only non-empty lines from ``Shell.output`` or
    ``Shell.errors`` will be returned. (Default: ``True``)
    """
    def __init__(self, has_input=False, record_output=True, record_errors=True,
                 strip_empty=True):
        self.has_input = has_input
        self.record_output = record_output
        self.record_errors = record_errors
        self.strip_empty = strip_empty

        self.last_command = ''
        self.line_breaks = '\n'
        self.pid = None
        self.code = 0
        self._popen = None
        self._stdout = ''
        self._stderr = ''

    def _split_command(self, command):
        """
        Splits a string command into the individual args to pass to ``Popen``.

        If the ``command`` is an array, it is left untouched.
        """
        if isinstance(command, (tuple, list)):
            return command

        return shlex.split(command)

    def _handle_output(self, stdout, stderr):
        """
        Takes stdout/stderr & optionally retains them internally.

        Requires a ``stdout`` parameter, which should either be the output as
        a string or ``None``.

        Requires a ``stderr`` parameter, which should either be the errors as
        a string or ``None``.

        Records nothing if the ``record_*`` options have been set to ``False``.
        """
        if self.record_output:
            if stdout is not None:
                self._stdout += stdout

        if self.record_errors:
            if stderr is not None:
                self._stderr += stderr

    def _communicate(self, the_input=None):
        """
        Handles communicating with the process & optionally sends input.

        Optionally accepts a ``the_input`` parameter, which can be a string
        to send to the process. (Default: ``None``)
        """
        stdout, stderr = self._popen.communicate(input=the_input)
        self._handle_output(stdout, stderr)

        if self._popen.returncode is not None:
            self.code = self._popen.returncode

    def run(self, command):
        """
        Runs a given command.

        Requires a ``command`` parameter should be either a string command
        (easier) or an array of arguments to send as the command (if you know
        what you're doing).

        Returns the ``Shell`` instance.

        Example::

            >>> from shell import Shell
            >>> sh = Shell()
            >>> sh.run('ls- alh')

        """
        self.last_command = command
        command_bits = self._split_command(command)
        kwargs = {
            'stdout': subprocess.PIPE,
            'stderr': subprocess.PIPE,
            'universal_newlines': True,
        }

        if self.has_input:
            kwargs['stdin'] = subprocess.PIPE

        self._popen = subprocess.Popen(
            command_bits,
            **kwargs
        )
        self.pid = self._popen.pid

        if not self.has_input:
            self._communicate()

        return self

    def write(self, the_input):
        """
        If you're working with an interactive process, sends that input to
        the process.

        This needs to be used in conjunction with the ``has_input=True``
        parameter.

        Requires a ``the_input`` parameter, which should be a string of the
        input to send to the command.

        Returns the ``Shell`` instance.

        Example::

            >>> from shell import Shell
            >>> sh = Shell(has_input=True)
            >>> sh.run('cat -u')
            >>> sh.write('Hello world!')

        """
        if not self._popen:
            raise MissingCommandException(
                "No command has been provided. Please call ``run`` first."
            )

        self._communicate(the_input)
        return self

    def kill(self):
        """
        Kills a given process.

        Example::

            >>> from shell import Shell
            >>> sh = Shell()
            >>> sh.run('some_long_running_thing')
            >>> sh.kill()

        """
        if not self._popen:
            raise MissingCommandException(
                "No command has been provided. Please call ``run`` first."
            )

        self._popen.kill()
        return self

    def output(self, raw=False):
        """
        Returns the output from running a command.

        Optionally accepts a ``raw`` parameter, which should be a boolean. If
        ``raw`` is set to ``False``, you get an array of lines of output. If
        ``raw`` is set to ``True``, the raw string of output is returned.
        (Default: ``False``)

        Example::

            >>> from shell import Shell
            >>> sh = Shell()
            >>> sh.run('ls ~')
            >>> sh.output()
            [
                'hello.txt',
                'world.txt',
            ]

        """
        if raw:
            return self._stdout

        lines = self._stdout.split(self.line_breaks)

        if self.strip_empty:
            lines = [line for line in lines if line]

        return lines

    def errors(self, raw=False):
        """
        Returns the errors from running a command.

        Optionally accepts a ``raw`` parameter, which should be a boolean. If
        ``raw`` is set to ``False``, you get an array of lines of errors. If
        ``raw`` is set to ``True``, the raw string of errors is returned.
        (Default: ``False``)

        Example::

            >>> from shell import Shell
            >>> sh = Shell()
            >>> sh.run('ls /there-s-no-way-anyone/has/this/directory/please')
            >>> sh.errors()
            [
                'ls /there-s-no-way-anyone/has/this/directory/please: No such
                 file or directory'
            ]

        """
        if raw:
            return self._stderr

        lines = self._stderr.split(self.line_breaks)

        if self.strip_empty:
            lines = [line for line in lines if line]

        return lines


def shell(command, has_input=False, record_output=True, record_errors=True,
          strip_empty=True):
    """
    A convenient shortcut for running commands.

    Requires a ``command`` parameter should be either a string command
    (easier) or an array of arguments to send as the command (if you know
    what you're doing).

    Optionally accepts a ``has_input`` parameter, which should be a boolean.
    If set to ``True``, the command will wait to execute until you call the
    ``Shell.write`` method & send input. (Default: ``False``)

    Optionally accepts a ``record_output`` parameter, which should be a
    boolean.
    If set to ``True``, the stdout from the command will be recorded.
    (Default: ``True``)

    Optionally accepts a ``record_errors`` parameter, which should be a
    boolean.
    If set to ``True``, the stderr from the command will be recorded.
    (Default: ``True``)

    Optionally accepts a ``strip_empty`` parameter, which should be a boolean.
    If set to ``True``, only non-empty lines from ``Shell.output`` or
    ``Shell.errors`` will be returned. (Default: ``True``)

    Returns the ``Shell`` instance, which has been run with the given command.

    Example::

        >>> from shell import shell
        >>> sh = shell('ls -alh *py')
        >>> sh.output()
        ['hello.py', 'world.py']

    """
    sh = Shell(
        has_input=has_input,
        record_output=record_output,
        record_errors=record_errors,
        strip_empty=strip_empty
    )
    return sh.run(command)



# Copyright 2014-2016 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
import logging
from subprocess import Popen, PIPE, TimeoutExpired

from urwid import (AttrMap, Columns, Divider, Filler, Overlay,
                   GridFlow, Frame, Padding, Pile, Text, WidgetWrap)

from ubuntui.widgets.buttons import PlainButton, MenuSelectButton
from ubuntui.views import InfoDialogWidget
from ubuntui.widgets import MetaScroll

from bundleplacer.charmstore_api import MetadataController
from bundleplacer.ui.charmstore import CharmstoreColumn, CharmStoreSearchWidget
from bundleplacer.ui.filter_box import FilterBox
from bundleplacer.ui.services_column import ServicesColumn
from bundleplacer.ui.machines_column import MachinesColumn
from bundleplacer.ui.relations_column import RelationsColumn

log = logging.getLogger('bundleplacer')


BUTTON_SIZE = 20


class UIState(Enum):
    PLACEMENT_EDITOR = 0
    RELATION_EDITOR = 1
    CHARMSTORE_VIEW = 2         # This is the default


class PlacementView(WidgetWrap):

    """
    Handles display of machines and services.

    displays nothing if self.controller is not set.
    set it to a PlacementController.

    :param do_deploy_cb: deploy callback from controller
    """

    def __init__(self, display_controller, placement_controller,
                 config, do_deploy_cb,
                 initial_state=UIState.CHARMSTORE_VIEW,
                 has_maas=False):
        self.display_controller = display_controller
        self.placement_controller = placement_controller
        self.config = config
        self.do_deploy_cb = do_deploy_cb
        self.state = initial_state
        self.has_maas = has_maas
        self.prev_state = None
        self.metadata_controller = MetadataController(placement_controller,
                                                      config)
        w = self.build_widgets()
        super().__init__(w)
        self.reset_selections(top=True)  # calls self.update

    def scroll_down(self):
        pass

    def scroll_up(self):
        pass

    def focus_footer(self):
        self.frame.focus_position = 'footer'
        self.footer_grid.focus_position = 1

    def handle_tab(self, backward):
        if self.state == UIState.RELATION_EDITOR:
            # Relation editor has no header col.
            tabloop = ['headercol1', 'col1', 'col2', 'footer']
        else:
            tabloop = ['headercol1', 'col1', 'headercol2', 'col2', 'footer']

        if not self.has_maas:
            tabloop.remove('headercol1')

        def goto_header_col1():
            self.frame.focus_position = 'header'
            self.header_columns.focus_position = 0

        def goto_header_col2():
            self.frame.focus_position = 'header'
            self.header_columns.focus_position = 1

        def goto_col1():
            self.frame.focus_position = 'body'
            self.columns.focus_position = 0

        def goto_col2():
            self.frame.focus_position = 'body'
            if self.state == UIState.PLACEMENT_EDITOR:
                self.focus_machines_column()
            elif self.state == UIState.RELATION_EDITOR:
                self.focus_relations_column()
            else:
                self.focus_charmstore_column()

        actions = {'headercol1': goto_header_col1,
                   'headercol2': goto_header_col2,
                   'col1': goto_col1,
                   'col2': goto_col2,
                   'footer': self.focus_footer}

        if self.frame.focus_position == 'header':
            cur = ['headercol1',
                   'headercol2'][self.header_columns.focus_position]
        elif self.frame.focus_position == 'footer':
            cur = 'footer'
        else:
            cur = ['col1', 'col2'][self.columns.focus_position]

        cur_idx = tabloop.index(cur)

        if backward:
            next_idx = cur_idx - 1
        else:
            next_idx = (cur_idx + 1) % len(tabloop)

        actions[tabloop[next_idx]]()

    def keypress(self, size, key):
        if key in ['tab', 'shift tab']:
            self.handle_tab('shift' in key)
            return key
        else:
            return self._w.keypress(size, key)

    def get_services_header(self):
        b = PlainButton("Clear All Placements",
                        on_press=self.do_clear_all)
        self.clear_all_button = AttrMap(b,
                                        'button_secondary',
                                        'button_secondary focus')

        self.services_buttons = [self.clear_all_button]
        self.services_button_grid = GridFlow(self.services_buttons,
                                             36, 1, 0, 'center')

        ws = [Divider(), Text(("body", "Services"), align='center'),
              Divider()]
        if self.has_maas:
            ws.append(self.services_button_grid)

        return Pile(ws)

    def get_charmstore_header(self, charmstore_column):
        series = self.placement_controller.bundle.series
        self.charm_search_widget = CharmStoreSearchWidget(self.do_add_charm,
                                                          charmstore_column,
                                                          self.config,
                                                          series)
        self.charm_search_header_pile = Pile([Divider(),
                                              Text(("body", "Add Charms"),
                                                   align='center'),
                                              Divider(),
                                              self.charm_search_widget])

        return self.charm_search_header_pile

    def get_machines_header(self, machines_column):
        b = PlainButton("Open in Browser",
                        on_press=self.browse_maas)
        self.open_maas_button = AttrMap(b,
                                        'button_secondary',
                                        'button_secondary focus')
        self.maastitle = Text("Connected to MAAS")
        maastitle_widgets = Padding(Columns([self.maastitle,
                                             (22, self.open_maas_button)]),
                                    align='center',
                                    width='pack', left=2,
                                    right=2)

        f = machines_column.machines_list.handle_filter_change
        self.filter_edit_box = FilterBox(f)
        pl = [Divider(),
              Text(('body',
                    "Ready Machines {}".format(MetaScroll().get_text()[0])),
                   align='center'),
              Divider(),
              maastitle_widgets,
              Divider(),
              self.filter_edit_box]

        self.machines_header_pile = Pile(pl)
        return self.machines_header_pile

    def update_machines_header(self):
        maasinfo = self.placement_controller.maasinfo
        maasname = "'{}' <{}>".format(maasinfo['server_name'],
                                      maasinfo['server_hostname'])
        self.maastitle.set_text("Connected to MAAS {}".format(maasname))

    def get_relations_header(self):
        return Pile([Divider(),
                     Text(('body', "Relation Editor"),
                          align='center'),
                     Divider()])

    def build_widgets(self):

        self.services_column = ServicesColumn(self.display_controller,
                                              self.placement_controller,
                                              self)

        self.machines_column = MachinesColumn(self.display_controller,
                                              self.placement_controller,
                                              self)
        self.relations_column = RelationsColumn(self.display_controller,
                                                self.placement_controller,
                                                self,
                                                self.metadata_controller)
        self.charmstore_column = CharmstoreColumn(self.display_controller,
                                                  self.placement_controller,
                                                  self,
                                                  self.metadata_controller)

        self.machines_header = self.get_machines_header(self.machines_column)
        self.relations_header = self.get_relations_header()
        self.services_header = self.get_services_header()
        self.charmstore_header = self.get_charmstore_header(
            self.charmstore_column)

        cs = [self.services_header, self.charmstore_header]

        self.header_columns = Columns(cs, dividechars=2)

        self.columns = Columns([self.services_column,
                                self.machines_column], dividechars=2)

        self.deploy_button = MenuSelectButton("\nCommit\n",
                                              on_press=self.do_deploy)
        self.deploy_button_label = Text("Some charms use default")
        self.placement_edit_body = Filler(Padding(self.columns,
                                                  align='center',
                                                  width=('relative', 95)),
                                          valign='top')
        b = AttrMap(self.deploy_button,
                    'frame_header',
                    'button_primary focus')
        self.footer_grid = GridFlow([self.deploy_button_label,
                                     Padding(b, width=28,
                                             align='center')],
                                    28, 1, 1, 'right')
        f = AttrMap(self.footer_grid,
                    'frame_footer',
                    'frame_footer')

        self.frame = Frame(header=self.header_columns,
                           body=self.placement_edit_body,
                           footer=f)
        return self.frame

    def update(self):
        if self.prev_state != self.state:
            h_opts = self.header_columns.options()
            c_opts = self.columns.options()

            if self.state == UIState.PLACEMENT_EDITOR:
                self.update_machines_header()
                self.header_columns.contents[-1] = (self.machines_header,
                                                    h_opts)
                self.columns.contents[-1] = (self.machines_column, c_opts)

            elif self.state == UIState.RELATION_EDITOR:
                self.header_columns.contents[-1] = (self.relations_header,
                                                    h_opts)
                self.columns.contents[-1] = (self.relations_column, h_opts)
            elif self.state == UIState.CHARMSTORE_VIEW:
                self.header_columns.contents[-1] = (self.charmstore_header,
                                                    h_opts)
                self.columns.contents[-1] = (self.charmstore_column, h_opts)

            self.prev_state = self.state

        self.services_column.update()

        if self.state == UIState.PLACEMENT_EDITOR:
            self.machines_column.update()
        elif self.state == UIState.RELATION_EDITOR:
            self.relations_column.update()

        unplaced = self.placement_controller.unassigned_undeployed_services()
        all = self.placement_controller.services()
        n_subs = len([c for c in all if c.subordinate])
        n_total = len(all) - n_subs
        remaining = len(unplaced) - n_subs
        if remaining > 0:
            dmsg = "\nAuto-assigning {}/{} services".format(remaining,
                                                            n_total)
        else:
            dmsg = ""
        self.deploy_button_label.set_text(dmsg)

    def browse_maas(self, sender):

        bc = self.config.juju_env['bootstrap-config']
        try:
            p = Popen(["sensible-browser", bc['maas-server']],
                      stdout=PIPE, stderr=PIPE)
            outs, errs = p.communicate(timeout=5)

        except TimeoutExpired:
            # went five seconds without an error, so we assume it's
            # OK. Don't kill it, just let it go:
            return
        e = errs.decode('utf-8')
        msg = "Error opening '{}' in a browser:\n{}".format(bc['name'], e)

        w = InfoDialogWidget(msg, self.remove_overlay)
        self.show_overlay(w)

    def do_clear_all(self, sender):
        self.placement_controller.clear_all_assignments()

    def do_add_charm(self, charm_name, charm_dict):
        """Add new service and focus its widget.

        """
        assert(self.state == UIState.CHARMSTORE_VIEW)
        service_name = self.placement_controller.add_new_service(charm_name,
                                                                 charm_dict)
        self.frame.focus_position = 'body'
        self.columns.focus_position = 0
        self.relations_column.add_charm(charm_name)
        self.update()
        self.services_column.select_service(service_name)

    def do_add_bundle(self, bundle_dict):
        assert(self.state == UIState.CHARMSTORE_VIEW)
        new_bundle = self.placement_controller.merge_bundle(bundle_dict)
        self.frame.focus_position = 'body'
        self.columns.focus_position = 0
        charms = list(set([s.charm_name for s in new_bundle.services]))
        for charm in charms:
            self.relations_column.add_charm(charm)
        self.update()
        first_service = new_bundle.services[0].service_name
        self.services_column.select_service(first_service)

    def do_clear_machine(self, sender, machine):
        self.placement_controller.clear_assignments(machine)

    def clear_selections(self):
        self.services_column.clear_selections()
        self.machines_column.clear_selections()

    def reset_selections(self, top=False):
        self.clear_selections()
        self.state = UIState.CHARMSTORE_VIEW
        self.update()
        self.columns.focus_position = 0

        if top:
            self.services_column.focus_top()
        else:
            self.services_column.focus_next()

    def focus_machines_column(self):
        self.columns.focus_position = 1
        self.machines_column.focus_prev_or_top()

    def focus_relations_column(self):
        self.columns.focus_position = 1
        self.relations_column.focus_prev_or_top()

    def focus_charmstore_column(self):
        self.columns.focus_position = 1
        self.charmstore_column.focus_prev_or_top()

    def edit_placement(self):
        self.state = UIState.PLACEMENT_EDITOR
        self.update()
        self.focus_machines_column()

    def show_default_view(self):
        self.state = UIState.CHARMSTORE_VIEW
        self.update()

    def edit_relations(self, service):
        self.state = UIState.RELATION_EDITOR
        self.relations_column.set_service(service)
        self.update()
        self.focus_relations_column()

    def do_deploy(self, sender):
        self.do_deploy_cb()

    def show_overlay(self, overlay_widget):
        self.orig_w = self._w
        self._w = Overlay(top_w=overlay_widget,
                          bottom_w=self._w,
                          align='center',
                          width=('relative', 60),
                          min_width=80,
                          valign='middle',
                          height='pack')

    def remove_overlay(self, overlay_widget):
        # urwid note: we could also get orig_w as
        # self._w.contents[0][0], but this is clearer:
        self._w = self.orig_w



from urwid import (WidgetWrap, Pile,
                   Filler)
from ubuntui.utils import Color, Padding
from ubuntui.widgets.hr import HR
from ubuntui.widgets.text import Instruction
from ubuntui.widgets.buttons import cancel_btn, menu_btn, PlainButton
from conjure.models.bundle import BundleModel


class CloudView(WidgetWrap):
    def __init__(self, app, clouds, cb):
        self.app = app
        self.cb = cb
        self.clouds = clouds
        self.config = self.app.config
        super().__init__(self._build_widget())

    def keypress(self, size, key):
        if key in ['tab', 'shift tab']:
            self._swap_focus()
        return super().keypress(size, key)

    def _swap_focus(self):
        w_count = len(self._w.body.contents) - 1
        if self._w.body.focus_position >= 2 and \
           self._w.body.focus_position < w_count:
            self._w.body.focus_position = w_count
        else:
            self._w.body.focus_position = 2

    def _build_buttons(self):
        cancel = cancel_btn(on_press=self.cancel)
        buttons = [
            Padding.line_break(""),
            Color.button_secondary(cancel,
                                   focus_map='button_secondary focus'),
        ]
        return Pile(buttons)

    def _build_widget(self):
        total_items = [
            Padding.center_60(Instruction("Clouds")),
            Padding.center_60(HR())
        ]
        for item in self.clouds:
            # Skip MAAS as this is a special case to be handled
            # a bit differently.
            if item == "maas":
                continue
            total_items.append(Padding.center_60(
                Color.body(
                    menu_btn(label=item,
                             on_press=self.submit),
                    focus_map='menu_button focus'
                )
            ))

        total_items.append(Padding.line_break(""))
        total_items.append(Padding.center_50(
            Color.body(
                PlainButton(label="Connect to an existing MAAS",
                            on_press=self.submit_new_maas),
                focus_map='menu_button focus')
        ))

        total_items.append(
            Padding.center_60(HR()))
        total_items.append(Padding.center_20(self._build_buttons()))
        return Filler(Pile(total_items), valign='middle')

    def submit(self, result):
        self.cb(result.label, create_cloud=True)

    def submit_new_maas(self, btn):
        return self.app.controllers['newcloud'].render("maas")

    def cancel(self, btn):
        self.cb(back=True)



#
# client.py - Client routines for MAAS API
#
# Copyright 2014 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import bson
from requests_oauthlib import OAuth1
import requests
import json
from urllib.parse import urlsplit


class MaasClient:

    """ Client Class
    """

    def __init__(self, auth):
        """ Entry point to client routines for interfacing
        with MAAS api.

        :param auth: MAAS Authorization class (required)
        """
        self.auth = auth
        self.server_hostname = urlsplit(auth.api_url).netloc

    def _oauth(self):
        """ Generates OAuth attributes for protected resources

        :returns: OAuth class
        """
        oauth = OAuth1(self.auth.consumer_key,
                       client_secret=self.auth.consumer_secret,
                       resource_owner_key=self.auth.token_key,
                       resource_owner_secret=self.auth.token_secret,
                       signature_method='PLAINTEXT',
                       signature_type='query')
        return oauth

    def get(self, url, params=None):
        """ Performs a authenticated GET against a MAAS endpoint

        :param url: MAAS endpoint
        :param params: extra data sent with the HTTP request
        """
        return requests.get(url=self.auth.api_url + url,
                            auth=self._oauth(),
                            params=params)

    def post(self, url, params=None):
        """ Performs a authenticated POST against a MAAS endpoint

        :param url: MAAS endpoint
        :param params: extra data sent with the HTTP request
        """
        return requests.post(url=self.auth.api_url + url,
                             auth=self._oauth(),
                             data=params)

    def put(self, url, params=None):
        """ Performs a authenticated PUT against a MAAS endpoint

        :param url: MAAS endpoint
        :param params: extra data sent with the HTTP request
        """
        return requests.put(url=self.auth.api_url + url,
                            auth=self._oauth(),
                            data=params)

    def delete(self, url, params=None):
        """ Performs a authenticated DELETE against a MAAS endpoint

        :param url: MAAS endpoint
        :param params: extra data sent with the HTTP request
        """
        return requests.delete(url=self.auth.api_url + url,
                               auth=self._oauth())

    def get_server_config(self, param):
        """ Query maas server config.
        see MAAS docs for available params
        """
        params = {}
        params['op'] = 'get_config'
        params['name'] = param
        res = self.get('/maas/', params)
        if not res.ok:
            return None
        return res.json()

    ###########################################################################
    # Boot Images API
    ###########################################################################
    def boot_images(self, uuid):
        """ Query boot images list

        :param str uuid: uuid of cluster
        :returns: list of boot images
        :rtype: list
        """
        _url = "/nodegroups/{uuid}/boot-images/".format(uuid=uuid)
        res = self.get(_url)
        if res.ok:
            return json.loads(res.text)
        return []

    def import_boot_images(self):
        """ Import boot images on all accepted controllers

        :returns: true on success, false on failure
        :rtype: bool
        """
        _url = "/nodegroups/"
        res = self.post(_url, dict(op='import_boot_images'))
        if res.ok:
            return True
        return False

    def report_boot_images(self, uuid):
        """ Describe imported images

        :param str uuid: uuid of cluster
        :returns: information on import images
        :rtype: dict
        """
        _url = "/nodegroups/{uuid}/boot-images".format(uuid=uuid)
        res = self.post(_url, dict(op='report_boot_images'))
        if res.ok:
            return json.loads(res.text)
        return {}

    ###########################################################################
    # Node API
    ###########################################################################
    @property
    def nodes(self):
        """ Nodes managed by MAAS

        :returns: managed nodes
        :rtype: list
        """
        res = self.get('/nodes/', dict(op='list'))
        if res.ok:
            return json.loads(res.text)
        return []

    def nodes_V2(self, **params):
        """ Nodes managed by MAAS

        See http://maas.ubuntu.com/docs/api.html#nodes

        :param params: keyword parameters to filter returned nodes
                       allowed values include hostnames, mac_addresses,
                       zone.
        :returns: managed nodes
        :rtype: list
        """
        params['op'] = 'list'
        res = self.get('/nodes/', params)
        if res.ok:
            machines = map(Machine, res.json())
            # Filter by state post response manually till maas fixes.
            if 'state' in params:
                machines = [m for m in machines if m.state == params['state']]
            return machines
        return []

    def node_get(self, node_id):
        res = self.get('/nodes/%s' % node_id)
        if res.ok:
            return Machine(res.json())
        return None

    def node_acquire(self, **params):
        """
        See http://maas.ubuntu.com/docs/api.html#nodes
        """
        params['op'] = 'acquire'
        res = self.post('/nodes/', params)
        if res.ok:
            return Machine(res.json())
        return []

    def node_release(self, node_id):
        """ Release a node back into the pool.
        """
        res = self.post('/nodes/%s/' % node_id, {'op': 'release'})
        return res.ok

    def node_start(self, node_id, user_data=None, distro_series=None):
        """ Power up a node

        :param node_id: machine identification
        :returns: True on success False on failure
        """
        params = {'op': 'start'}
        if user_data:
            params['user_data'] = user_data
        if distro_series:
            params['distro_series'] = distro_series
        res = self.post('/nodes/%s/' % node_id, params)
        return res.ok

    def node_stop(self, node_id):
        """ Shutdown a node

        :param node_id: machine identification
        :returns: True on success False on failure
        """
        res = self.post('/nodes/%s/' % node_id, {'op': 'stop'})
        return res.ok

    def node_details(self, system_id):
        """ Node Details from e.g. lshw.
        May be encoded as XML.

        :returns: dictionary of commissioning details
        :rtype: dict
        """
        res = self.get('/nodes/{}/'.format(system_id),
                       dict(op='details'))
        if res.ok:
            ds = bson.decode_all(res.content)
            if len(ds) == 0:
                return None
            return ds[0]
        return None

    def nodes_accept_all(self):
        """ Accept all commissioned nodes

        :returns: Status
        :rtype: bool
        """

        res = self.post('/nodes/', dict(op='accept_all'))
        if res.ok:
            return True
        return False

    def node_commission(self, system_id):
        """ (Re)commission a node

        :param system_id: machine identification
        :returns: True on success False on failure
        """
        res = self.post('/nodes/%s' % (system_id,), dict(op='commission'))
        if res.ok:
            return True
        return False

    def node_remove(self, system_id):
        """ Delete a node

        :param system_id: machine identification
        :returns: True and success False on failure
        """
        res = self.delete('/nodes/%s' % (system_id,))
        if res.ok:
            return True
        return False

    ###########################################################################
    # Nodegroups API
    ###########################################################################
    @property
    def nodegroups(self):
        """ List nodegroups

        :returns: List of nodegroups
        :rtype list:
        """
        res = self.get('/nodegroups/', dict(op='list'))
        if res.ok:
            return json.loads(res.text)
        return []

    def nodegroups_download_progress(self, uuid):
        """ Report download progress for a cluster controller

        :param str uuid: uuid of controller
        :returns: file information
        :rtype: dict
        """
        _url = "/nodegroups/{uuid}/".format(uuid=uuid)
        res = self.post(_url, dict(op='report_download_progress'))
        if res.ok:
            return json.loads(res.text)
        return {}

    ###########################################################################
    # Tag API
    ###########################################################################
    @property
    def tags(self):
        """ List tags known to MAAS

        :returns: List of tags or empty list
        """
        res = self.get('/tags/', dict(op='list'))
        if res.ok:
            return json.loads(res.text)
        return []

    def tag_new(self, tag):
        """ Create tag if it doesn't exist.

        :param tag: Tag name
        :returns: Success/Fail boolean
        """
        tags = {tagmd['name'] for tagmd in self.tags}
        if tag not in tags:
            res = self.post('/tags/', dict(op='new', name=tag))
            return res.ok
        return False

    def tag_delete(self, tag):
        """ Delete a tag

        :param tag: tag id
        :type tag: str
        :returns: True on success False on failure
        :rtype: bool
        """
        res = self.delete('/tags/%s' % (tag,))
        if res.ok:
            return True
        return False

    def tag_machine(self, tag, system_id):
        """ Tag the machine with the specified tag.

        :param tag: Tag name
        :type tag: str
        :param system_id: ID of node
        :type system_id: str
        :returns: Success or Fail
        :rtype: bool
        """

        # Make use of rest api
        res = self.post('/tags/%s/' % (tag,),
                        dict(op='update_nodes',
                             add=system_id))
        if res.ok:
            return True
        return False

    def tag_name(self, nodes):
        """ Tag each managed node with its hostname.

        This is a bit ugly. Since we want to be able to juju deploy to
        a particular node that the user has selected, we use juju's
        constraints support for maas. Unfortunately, juju didn't
        implement maas-name directly, we have to tag each node with
        its hostname for now so that we can pass that tag as a
        constraint to juju.

        """
        for machine in nodes:
            system_id = machine['system_id']
            if 'tag_names' not in machine['tag_names'] or \
               system_id not in machine['tag_names']:
                self.tag_new(system_id)
                self.tag_machine(system_id, system_id)

    def tag_fpi(self, nodes):
        """ Tag each DECLARED host with the FPI tag.

        Also a little strange: we could define a tag with
        'definition=true()' and automatically tag each node. However,
        each time we un-tag a node, maas evaluates the xpath
        expression again and re-tags it. So, we do it once, manually,
        when the machine is in the DECLARED state (also to avoid
        re-tagging things that have already been tagged).

        :param maas: MAAS object representing all managed nodes
        """
        FPI_TAG = 'use-fastpath-installer'
        self.tag_new(FPI_TAG)
        for machine in nodes:
            if machine['status'] == 0:
                self.tag_machine(FPI_TAG, machine['system_id'])

    ###########################################################################
    # Users API
    ###########################################################################
    @property
    def users(self):
        """ List users on MAAS

        :returns: List of registered users or an empty list
        """
        res = self.get('/users/')
        if res.ok:
            return json.loads(res.text)
        return []

    ##########################################################################
    # Network API
    ##########################################################################
    @property
    def networks(self):
        """ List networks
        :returns: List of networks
        """
        res = self.get('/networks/')
        if res.ok:
            return res.json()
        return []

    ###########################################################################
    # Zone API
    ###########################################################################
    @property
    def zones(self):
        """ List physical zones

        :returns: List of managed zones or empty list
        """
        res = self.get('/zones/')
        if res.ok:
            return json.loads(res.text)
        return []

    def zone_new(self, name, description="Zone created by API"):
        """ Create a physical zone

        :param name: Name of the zone
        :param description: Description of zone.
        :returns: True on success False on failure
        """
        res = self.post('/zones/', dict(name=name,
                                        description=description))
        if res.ok:
            return True
        return False

    def zone_delete(self, name):
        """ Delete a zone.

        :param name: Name of the zone to be deleted
        :returns: True on success & False on failure.
        """
        res = self.delete('/zones/{}/'.format(name))
        return res.ok


class vocab(dict):
    __slots__ = ()

    def __getattr__(self, k):
        try:
            return self[k]
        except KeyError:
            return AttributeError(k)

    def label(self, value):
        for k, v in self.items():
            if v == value:
                return k

MAAS_STATES = vocab(
    DECLARED=0,
    COMMISSIONING=1,
    FAILED_TESTS=2,
    MISSING=3,
    READY=4,
    RESERVED=5,
    ALLOCATED=6,
    RETIRED=7)


class Machine(dict):

    __slots__ = ()

    @property
    def hostname(self):
        return self['hostname']

    @property
    def arch(self):
        return self['architecture']

    @property
    def status(self):
        return self['status']

    @property
    def cpu_cores(self):
        return self['cpu_count']

    @property
    def mem(self):
        """ Size Memory in mb"""
        return self['memory']

    @property
    def disk(self):
        """ Size root disk in gb"""
        return self['storage']

    @property
    def system_id(self):
        return self['system_id']

    @property
    def tags(self):
        return self.get('tags', [])

    @property
    def ip_addresses(self):
        return self['ip_addresses']

    @property
    def mac_addresses(self):
        return [m['mac_address'] for m in self.get('macaddress_set', [])]

    @property
    def status_label(self):
        return MAAS_STATES.label(self.status)



from jinja2 import FileSystemLoader, Environment
from conjure.utils import spew
from tempfile import NamedTemporaryFile
import yaml


def render_charm_conf(name, options):
    """ Render a yaml config suitable for charm deployment

    Arguments:
    name: service/charm name
    options: dictionary of config options and their values

    Returns:
    Path to charm config file
    """
    ctx = dict(name=options)
    with NamedTemporaryFile(mode='w+', encoding='utf-8') as tempf:
        tempf.write(yaml.dump(ctx, default_flow_style=False))
        return tempf.name


def load(name, path):
    """ load template file
    :param str name: name of template file
    :param str path: directory location of templates
    """
    env = Environment(
        loader=FileSystemLoader(path))
    return env.get_template(name)


def save(template, opts):
    """ Saves template to temporary file

    Arguments:
    template: loaded jinja template
    opts: dictionary of items to be passed into template
    """
    modified = template.render(**opts)
    with NamedTemporaryFile(mode='w', encoding='utf-8',
                            delete=False) as tempf:
        spew(tempf.name, modified)
        return tempf.name



class MacumbaError(Exception):

    "Base error class"


class CharmNotFoundError(MacumbaError):

    "Error when getting charm store url"


class LoginError(MacumbaError):

    "Error logging in to juju api"


class ServerError(MacumbaError):

    "Generic error response from server"

    def __init__(self, message, response):
        self.response = response
        super().__init__(self, message)


class BadResponseError(MacumbaError):

    "Unable to parse response from server"


class ConnectionClosedError(MacumbaError):

    "Attempted to receive messages from closed connection"


class UnknownRequestError(MacumbaError):

    "Attempted to receive a message with an unknown ID"


class RequestTimeout(MacumbaError):

    "Request timed out"



from conjure.ui.views.cloud import CloudView
from conjure.juju import Juju
from conjure.models.bundle import BundleModel
from conjure.utils import pollinate


class CloudController:

    def __init__(self, app):
        self.app = app

    def _list_clouds(self):
        """ Returns list of clouds filtering out any results
        """
        clouds = set(Juju.clouds().keys())

        if BundleModel.whitelist():
            whitelist = set(BundleModel.whitelist())
            return sorted(list(clouds & whitelist))

        elif BundleModel.blacklist():
            blacklist = set(BundleModel.blacklist())
            return sorted(list(clouds ^ blacklist))

        return sorted(list(clouds))

    def finish(self, cloud=None, create_cloud=False, back=False):
        """ Load the Model controller passing along the selected cloud.

        Arguments:
        cloud: Cloud to create the controller/model on.
        create_cloud: True/False, if true display create cloud interface
        back: if true loads previous controller
        """
        if back:
            return self.app.controllers['welcome'].render()

        if create_cloud:
            return self.app.controllers['newcloud'].render(cloud)

        pollinate(self.app.session_id, 'CS', self.app.log)

    def render(self):
        self.clouds = self._list_clouds()
        self.config = self.app.config
        self.excerpt = ("Please select from a list of available clouds")
        self.view = CloudView(self.app,
                              self.clouds,
                              self.finish)

        self.app.ui.set_header(
            title="Cloud Providers",
            excerpt=self.excerpt
        )
        self.app.ui.set_body(self.view)



from urwid import WidgetWrap, Text, Filler, Pile
from ubuntui.utils import Padding, Color
from ubuntui.widgets.buttons import (cancel_btn)


class FinishView(WidgetWrap):
    def __init__(self, app, cb):
        self.app = app
        self.cb = cb
        self.text = Text("Installing bundle...")
        _pile = [
            Padding.center_79(self.text),
            Padding.line_break(""),
            Padding.center_20(self.buttons())
        ]
        super().__init__(Filler(Pile(_pile), valign="middle"))

    def buttons(self):
        cancel = cancel_btn(label="Quit", on_press=self.cancel)

        buttons = [
            Color.button_secondary(cancel, focus_map='button_secondary focus')
        ]
        return Pile(buttons)

    def set_status(self, msg):
        current = self.text.get_text()[0]
        self.text.set_text("{}\n{}".format(current, msg))

    def cancel(self, btn):
        self.cb()



# Copyright 2014-2016 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from urwid import Divider, Pile, Text, WidgetWrap

from bundleplacer.ui.machines_list import MachinesList


log = logging.getLogger('bundleplacer')


BUTTON_SIZE = 20


class MachinesColumn(WidgetWrap):

    """Shows machines or a link to MAAS to add more"""

    def __init__(self, display_controller, placement_controller,
                 placement_view):
        self.display_controller = display_controller
        self.placement_controller = placement_controller
        self.placement_view = placement_view
        w = self.build_widgets()
        super().__init__(w)
        self.update()

    def selectable(self):
        return True

    def build_widgets(self):
        self.machines_list = MachinesList(self.placement_controller,
                                          self.display_controller,
                                          None,
                                          show_hardware=True,
                                          show_assignments=False,
                                          show_placeholders=False,
                                          show_only_ready=True,
                                          title_widgets=[])

        self.machines_list.update()

        self.machines_list_pile = Pile([self.machines_list,
                                        Divider()])

        return self.machines_list_pile

    def update(self):
        self.machines_list.update()
        maasinfo = self.placement_controller.maasinfo
        empty_maas_msg = ("There are no available machines.\n"
                          "Open {} to add machines to "
                          "'{}':".format(maasinfo['server_name'],
                                         maasinfo['server_hostname']))

        self.empty_maas_widgets = Pile([Text([('error_icon',
                                               "\N{WARNING SIGN} "),
                                              empty_maas_msg],
                                             align='center')])

        # 2 machines is the subordinate placeholder + juju default:
        opts = self.machines_list_pile.options()
        if len(self.placement_controller.machines()) == 2:
            self.machines_list_pile.contents[0] = (self.empty_maas_widgets,
                                                   opts)
        else:
            self.machines_list_pile.contents[0] = (self.machines_list,
                                                   opts)

    def clear_selections(self):
        for mw in self.machines_list.machine_widgets:
            mw.is_selected = False

    def focus_prev_or_top(self):
        self.update()
        try:
            self.machines_list_pile.focus_position = 0
            self.machines_list.focus_prev_or_top()
        except IndexError:
            log.debug("caught indexerror in machinesColumn focus_prev_or_top")
            pass



#
# __init__.py - MAAS instance state
#
# Copyright 2014 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import Counter
from enum import Enum
import json
import logging
import os
from threading import RLock
import time

from maasclient.auth import MaasAuth
from maasclient import MaasClient

from bundleplacer.async import submit
from bundleplacer.machine import Machine
from bundleplacer.utils import human_to_mb


log = logging.getLogger('bundleplacer')


def satisfies(machine, constraints):
    """Evaluates whether a MAAS machine's hardware matches constraints.

    If constraints is None or an empty dict, then any machine will be
    evaluated as satisfying the constraints.

    .. note::

        That if a machine has '*' as a value, that value satisfies
        any constraint.

    If successful the return will be (True, [])

    :rtype: tuple
    :returns: (bool, [list-of-failed constraint keys])

    """
    kmap = dict(mem='memory',
                arch='architecture',
                storage='storage',
                cpu_cores='cpu_count')
    kmap['root-disk'] = 'storage'

    cons_checks = []

    if constraints is None:
        return (True, [])

    for k, v in constraints.items():
        if k == 'arch':
            mval = machine.machine[kmap[k]]
            if mval != '*' and mval != v:
                cons_checks.append(k)
        else:
            mval = machine.machine[kmap[k]]

            if mval == '*':
                # '*' always satisfies.
                continue

            if not str(v).isdecimal():
                v = human_to_mb(v)

            if mval < v:
                cons_checks.append(k)

    rval = (len(cons_checks) == 0), cons_checks
    return rval


class MaasMachineStatus(Enum):
    """Symbolic names for maas API status numbers.

    -1, UNKNOWN is never returned by maas API. It's used here to
    denote a MaasMachine object that wasn't created from a Maas API
    return.
    """
    UNKNOWN = -1
    NEW = 0
    COMMISSIONING = 1
    FAILED_COMMISSIONING = 2
    MISSING = 3
    READY = 4
    RESERVED = 5
    # as of maas 1.7, state #s 6, and 9-15 are mapped
    # onto 6 by the view that services the nodes/
    # url. so we will only ever see '6' for any of
    # these, until sometime in the future.
    DEPLOYED = 6
    RETIRED = 7
    BROKEN = 8
    DEPLOYING = 9  # see DEPLOYED
    MAAS_1_7_ALLOCATED = 10  # the actual "ALLOCATED" state. see DEPLOYED
    FAILED_DEPLOYMENT = 11  # see DEPLOYED.
    RELEASING = 12
    FAILED_RELEASING = 13
    DISK_ERASING = 14
    FAILED_DISK_ERASING = 15
    ALLOCATED = 6  # for backward compatibility with internal uses

    def __str__(self):
        return self.name.lower()


class MaasMachine(Machine):
    """ Single maas machine """

    def __eq__(self, other):
        return self.hostname == other.hostname

    def __hash__(self):
        return hash(self.hostname)

    @property
    def hostname(self):
        """ Query hostname reported by MaaS

        :returns: hostname
        :rtype: str
        """
        return self.machine.get('hostname', '')

    @property
    def status(self):
        """ Status of machine state

        :returns: status enum
        :rtype: MaasMachineStatus
        """
        return MaasMachineStatus(self.machine.get('status',
                                                  MaasMachineStatus.UNKNOWN))

    @property
    def zone(self):
        """ Zone information

        :returns: zone information
        :rtype: dict
        """
        return self.machine.get('zone', {})

    @property
    def cpu_cores(self):
        """ Returns number of cpu-cores

        :returns: number of cpus
        :rtype: str
        """
        return self.machine.get('cpu_count', '0')

    @property
    def storage(self):
        """ Return storage

        :returns: storage size
        :rtype: str
        """
        try:
            _storage_in_gb = int(self.machine.get('storage')) / 1024
        except ValueError:
            return "N/A"
        return "{size:.2f}G".format(size=_storage_in_gb)

    @property
    def arch(self):
        """ Return architecture

        :returns: architecture type
        :rtype: str
        """
        return self.machine.get('architecture')

    @property
    def mem(self):
        """ Return memory

        :returns: memory size
        :rtype: str
        """
        try:
            _mem = int(self.machine.get('memory'))
        except ValueError:
            return "N/A"
        if _mem > 1024:
            _mem = _mem / 1024
            return "{size}G".format(size=str(_mem))
        else:
            return "{size}M".format(size=str(_mem))

    @property
    def power_type(self):
        """ Machine power type

        :returns: machines power type
        :rtype: str
        """
        return self.machine.get('power_type', 'None')

    @property
    def instance_id(self):
        """ Returns instance-id of a machine

        :returns: instance-id of machine
        :rtype: str
        """
        return self.machine.get('resource_uri', '')

    @property
    def system_id(self):
        """ Returns system id of a maas machine

        :returns: system id of machine
        :rtype: str
        """
        return self.machine.get('system_id', '')

    @property
    def ip_addresses(self):
        """ Ip addresses for machine

        :returns: ip addresses
        :rtype: list
        """
        return self.machine.get('ip_addresses', [])

    @property
    def macaddress_set(self):
        """ Macaddress set of maas machine

        :returns: list of dict(mac_address, resource_uri)
        :rtype: list
        """
        return self.machine.get('macaddress_set', [])

    @property
    def tag_names(self):
        """ Tag names for machine

        :returns: tags associated with machine
        :rtype: list
        """
        return self.machine.get('tag_names', [])

    @property
    def tag(self):
        """ Machine tag

        :returns: tag defined
        :rtype: str
        """
        return self.machine.get('tag', '')

    @property
    def owner(self):
        """ Machine owner

        :returns: owner
        :rtype: str
        """
        return self.machine.get('owner', 'root')

    def __repr__(self):
        return "<MaasMachine({dns_name},{state},{mem}," \
            "{storage},{cpus})>".format(dns_name=self.hostname,
                                        state=self.agent_state,
                                        mem=self.mem,
                                        storage=self.storage,
                                        cpus=self.cpu_cores)

    def __str__(self):
        return repr(self)

    def filter_label(self):
        d = dict(dns_name=self.hostname,
                 arch=self.arch,
                 tag=self.tag,
                 mem=self.mem,
                 storage=self.storage,
                 cpus=self.cpu_cores)
        return ("hostname:{dns_name} tag:{tag} mem:{mem} arch:{arch}"
                "storage:{storage} cores:{cpus}").format(**d)


class MaasState:
    """ Represents global MaaS state """

    def __init__(self, maas_client):
        self.maas_client = maas_client
        self._maas_client_nodes = []
        self._filtered_nodes = []
        self._nodes_lock = RLock()
        self._nodes_future = None
        self._start_time = 0
        self.server_hostname = maas_client.server_hostname

    def get_server_config(self, param):
        return self.maas_client.get_server_config(param)

    def nodes(self, constraints=None):
        """ Cache MAAS nodes
        """
        elapsed_time = time.time() - self._start_time
        if elapsed_time <= 20:
            return self._filtered_nodes

        def _do_update():
            nodes = self.maas_client.nodes
            self.invalidate_nodes_cache()
            return nodes

        if self._nodes_future:
            if self._nodes_future.done():
                with self._nodes_lock:
                    self._maas_client_nodes = self._nodes_future.result()
                self._nodes_future = None
        else:
            self._nodes_future = submit(_do_update, lambda _: None)

        if constraints:
            self._filtered_nodes = self._filter_nodes(self._maas_client_nodes,
                                                      constraints)
        else:
            self._filtered_nodes = self._maas_client_nodes

        self._start_time = time.time()
        return self._filtered_nodes

    def nodes_uncached(self, constraints=None):
        if constraints:
            return self._filter_nodes(self.maas_client.nodes)
        else:
            return self.maas_client.nodes

    def _filter_nodes(self, nodes, constraints):
        assert constraints is not None

        cd = dict(x.split('=') for x in constraints.split(' '))
        arch = cd.get('arch', None)
        tagstr = cd.get('tags', None)
        satisfying_nodes = []
        for n in self._maas_client_nodes:
            if arch:
                n_arch = n['architecture'].split('/')[0]
                if n_arch != arch:
                    continue
            if tagstr:
                c_tags = set(cd['tags'].split(','))
                n_tags = set(n['tag_names'])
                if not c_tags.issubset(n_tags):
                    continue
            satisfying_nodes.append(n)

        return satisfying_nodes

    def invalidate_nodes_cache(self):
        """Force reload on next access"""
        self._start_time = 0

    def machine(self, instance_id):
        """ Return single machine state

        :param str instance_id: machine instance_id
        :returns: machine
        :rtype: bundleplacer.maas.MaasMachine
        """
        for m in self.machines():
            if m.instance_id == instance_id:
                return m
        return None

    def machines(self, state=None, constraints=None):
        """Maas Machines

        :param state

        :param str constraints: a juju style constraints string that
        we parse for arch and tags

        :returns: machines known to Maas, except for juju bootstrap
            machine, matching state type, or all if state=None

        :rtype: list of MaasMachine

        """
        nodes = [n for n in self.nodes(constraints)
                 if n['hostname'] != 'juju-bootstrap.maas']
        all_machines = [MaasMachine(-1, m) for m in nodes]
        if state:
            return [m for m in all_machines if m.status == state]
        else:
            return all_machines

    def machines_summary(self):
        """ Returns summary of known machines and their states.
        """
        log.debug("in summary, self.nodes is {}".format(self.nodes()))
        return Counter([MaasMachineStatus(m['status'])
                        for m in self.nodes()])


def connect_to_maas(creds=None):
    if creds:
        api_host = creds['api_host']
        api_url = 'http://{}/MAAS/api/1.0'.format(api_host)
        api_key = creds['api_key']
        auth = MaasAuth(api_url=api_url,
                        api_key=api_key)
    else:
        auth = MaasAuth()
        auth.get_api_key('root')
    maas = MaasClient(auth)
    maas_state = MaasState(maas)
    return maas, maas_state



from ubuntui.widgets.buttons import (confirm_btn, cancel_btn)
from ubuntui.widgets.text import Instruction
from ubuntui.utils import Color, Padding
from ubuntui.widgets.hr import HR
from urwid import (WidgetWrap, Pile, Text, Columns, Filler)
from collections import OrderedDict
from ubuntui.widgets.input import (StringEditor, YesNo)
import os

# Network format
#
# { key: (Widget, Help Text)}

NETWORK = OrderedDict([
    ('_USE_LXD_BRIDGE',
     (YesNo(),
      'Use a new bridge')),
    ('LXD_BRIDGE',
     (StringEditor(default='lxdbr0'),
      'Bridge name')),
    # ('_LXD_CONFILE',
    #  (StringEditor(),
    #   None)),
    ('LXD_DOMAIN',
     (StringEditor(default='lxd'),
      'DNS domain for the bridge')),
    ('LXD_IPV4_ADDR',
     (StringEditor(default='10.10.0.1'),
      'IPv4 Address (e.g. 10.10.0.1)')),
    ('LXD_IPV4_NETMASK',
     (StringEditor(default='255.255.255.0'),
      'IPv4 netmask (e.g. 255.255.255.0)')),
    ('LXD_IPV4_NETWORK',
     (StringEditor(default='10.10.0.1/24'),
      'IPv4 network (e.g. 10.10.0.1/24)')),
    ('LXD_IPV4_DHCP_RANGE',
     (StringEditor(default='10.10.0.2,10.10.0.254'),
      'IPv4 DHCP range (e.g. 10.10.0.2,10.10.0.254)')),
    ('LXD_IPV4_DHCP_MAX',
     (StringEditor(default='250'),
      'IPv4 DHCP number of hosts (e.g. 250)')),
    ('LXD_IPV4_NAT',
     (YesNo(),
      'NAT IPv4 traffic'))
    # Enable this at some point
    # ('_LXD_IPV6_ADDR',
    #  (StringEditor(),
    #   None)),
    # ('_LXD_IPV6_MASK',
    #  (StringEditor(),
    #   None)),
    # ('_LXD_IPV6_NETWORK',
    #  (StringEditor(),
    #   None)),
    # ('_LXD_IPV6_NAT',
    #  (YesNo(),
    #   None)),
    # ('_LXD_IPV6_PROXY',
    #  (YesNo(),
    #   None))
])


class LXDSetupView(WidgetWrap):
    def __init__(self, app, cb):
        self.app = app
        self.input_items = NETWORK
        self.cb = cb
        _pile = [
            # Padding.center_60(Instruction(
            #     "Enter LXD information:")),
            Padding.center_60(Instruction(
                "Please configure networking for LXD"
            )),
            Padding.center_60(HR()),
            Padding.center_60(self.build_info()),
            Padding.line_break(""),
            Padding.center_20(self.buttons())
        ]
        super().__init__(Filler(Pile(_pile), valign="middle"))

    def _swap_focus(self):
        if self._w.body.focus_position == 2:
            self._w.body.focus_position = 4
        else:
            self._w.body.focus_position = 2

    def keypress(self, size, key):
        if key in ['tab', 'shift tab']:
            self._swap_focus()
        return super().keypress(size, key)

    def buttons(self):
        confirm = confirm_btn(on_press=self.submit)
        cancel = cancel_btn(on_press=self.cancel)

        buttons = [
            Color.button_primary(confirm, focus_map='button_primary focus'),
            Color.button_secondary(cancel, focus_map='button_secondary focus')
        ]
        return Pile(buttons)

    def build_info(self):
        items = [
            Text("There was no LXD bridge found on your system "
                 "which usually means this is your first time running "
                 "LXD."),
            Padding.line_break(""),
            Text("If you wish to do so now pressing confirm will drop you out "
                 "of the installer and walk you through configuring your "
                 "network for LXD. Once complete the installer will "
                 "start again from the beginning where you can choose "
                 "to deploy the bundle via LXD.")
        ]
        return Pile(items)

    def build_inputs(self):
        # FIXME: Skip this for now as we're shelling out to
        # dpkg-reconfigure lxd :\ yay and such.
        items = []
        for k in self.input_items.keys():
            widget, help_text = self.input_items[k]
            if isinstance(widget.value, bool):
                widget.set_default('Yes', True)
            if k.startswith('_'):
                # Don't treat 'private' keys as input
                continue
            col = Columns(
                [
                    ('weight', 0.5, Text(help_text, align='right')),
                    Color.string_input(widget,
                                       focus_map='string_input focus')
                ], dividechars=1
            )
            items.append(col)
            items.append(Padding.line_break(""))
        return Pile(items)

    def cancel(self, btn):
        self.cb(back=True)

    def submit(self, result):
        # self.cb(self.input_items)
        os.execl("/usr/share/conjure/run-lxd-config",
                 "/usr/share/conjure/run-lxd-config",
                 self.app.config["bin"])



from conjure.ui.views.services import ServicesView
from ubuntui.ev import EventLoop
from conjure.juju import Juju
from functools import partial
from conjure import async
from conjure.models.bundle import BundleModel
from conjure.utils import pollinate
import os.path as path
import os
import json
from subprocess import check_output


class FinishController:

    def __init__(self, app):
        self.app = app
        self._post_exec_pollinate = False
        self._pre_exec_pollinate = False

    def handle_exception(self, tag, exc):
        pollinate(self.app.session_id, tag, self.app.log)
        self.app.ui.show_exception_message(exc)

    def handle_post_exception(self, exc):
        """ If an exception occurs in the post processing,
        log it but don't die
        """
        pollinate(self.app.session_id, 'E002', self.app.log)
        self.app.log.exception(exc)

    def handle_pre_exception(self, exc):
        """ If an exception occurs in the pre processing,
        log it but don't die
        """
        pollinate(self.app.session_id, 'E003', self.app.log)
        self.app.ui.show_exception_message(exc)

    def _pre_exec(self, *args):
        """ Executes a bundles pre processing script if exists
        """
        self.app.log.debug("pre_exec start: {}".format(args))
        self._pre_exec_sh = path.join('/usr/share/',
                                      self.app.config['name'],
                                      'bundles',
                                      BundleModel.key(),
                                      'pre.sh')
        if not path.isfile(self._pre_exec_sh) \
           or not os.access(self._pre_exec_sh, os.X_OK):
            self.app.log.debug(
                "Unable to execute: {}, skipping".format(self._pre_exec_sh))
            return self._deploy_bundle()
        self.app.ui.set_footer('Running pre-processing tasks...')
        if not self._pre_exec_pollinate:
            pollinate(self.app.session_id, 'XA', self.app.log)
            self._pre_exec_pollinate = True

        self.app.log.debug("pre_exec running {}".format(self._pre_exec_sh))

        try:
            future = async.submit(partial(check_output,
                                          self._pre_exec_sh,
                                          shell=True,
                                          env=self.app.env),
                                  partial(self.handle_exception,
                                          "E002"))
            future.add_done_callback(self._pre_exec_done)
        except Exception as e:
            self.handle_exception("E002", e)

    def _pre_exec_done(self, future):
        result = json.loads(future.result().decode('utf8'))
        self.app.log.debug("pre_exec_done: {}".format(result))
        if result['returnCode'] > 0:
            return self.handle_pre_exception(Exception(
                'There was an error during the pre processing phase.'))
        self._deploy_bundle()

    def _deploy_bundle(self):
        """ Performs the bootstrap in between processing scripts
        """
        self.app.log.debug("Deploying bundle: {}".format(self.bundle))
        self.app.ui.set_footer('Deploying bundle...')
        pollinate(self.app.session_id, 'DS', self.app.log)
        future = async.submit(
            partial(Juju.deploy_bundle, self.bundle),
            partial(self.handle_exception, "ED"))
        future.add_done_callback(self._deploy_bundle_done)

    def _deploy_bundle_done(self, future):
        result = future.result()
        self.app.log.debug("deploy_bundle_done: {}".format(result.output()))
        if result.code > 0:
            self.handle_exception("ED", Exception(
                'There was an error deploying the bundle: {}.'.format(
                    result.errors())))
            return
        self.app.ui.set_footer('Deploy completed, waiting for '
                               'all services to be available.')
        pollinate(self.app.session_id, 'DC', self.app.log)
        EventLoop.set_alarm_in(1, self._post_exec)

    def _post_exec(self, *args):
        """ Executes a bundles post processing script if exists
        """
        self._post_exec_sh = path.join('/usr/share/',
                                       self.app.config['name'],
                                       'bundles',
                                       BundleModel.key(),
                                       'post.sh')

        if not path.isfile(self._post_exec_sh) \
           or not os.access(self._post_exec_sh, os.X_OK):
            self.app.log.debug(
                "Unable to execute: {}, skipping".format(self._post_exec_sh))
            return

        if not self._post_exec_pollinate:
            # We dont want to keep pollinating since this routine could
            # run multiple times
            pollinate(self.app.session_id, 'XB', self.app.log)
            self._post_exec_pollinate = True

        self.app.log.debug("post_exec running: {}".format(self._post_exec_sh))
        future = async.submit(partial(check_output,
                                      self._post_exec_sh,
                                      shell=True,
                                      env=self.app.env),
                              self.handle_post_exception)
        future.add_done_callback(self._post_exec_done)

    def _post_exec_done(self, future):
        try:
            result = json.loads(future.result().decode('utf8'))
            self.app.log.debug("post_exec_done: {}".format(result))
            self.app.ui.set_footer(result['message'])
            if result['returnCode'] > 0 or not result['isComplete']:
                self.app.log.error(
                    'There was an error during the post processing '
                    'phase, retrying.')
                EventLoop.set_alarm_in(5, self._post_exec)
            else:
                # Stop post processing loop and restart view refresh
                EventLoop.remove_alarms()
                EventLoop.set_alarm_in(1, self.refresh)
        except Exception as e:
            self.app.log.error(e)
            self.handle_exception("E002", e)

    def refresh(self, *args):
        self.view.refresh_nodes()
        EventLoop.set_alarm_in(1, self.refresh)

    def render(self, bundle):
        """ Render services status view

        Arguments:
        bundle: modified bundle to deploy
        """
        self.bundle = bundle
        self.view = ServicesView(self.app)

        self.app.ui.set_header(
            title=self.app.config['summary']
        )
        self.app.ui.set_body(self.view)
        self.app.ui.set_subheader(
            'Deploy Status - (Q)uit || UP/DOWN to Scroll')

        if not self.app.argv.status_only:
            EventLoop.set_alarm_in(1, self._pre_exec)
        EventLoop.set_alarm_in(1, self.refresh)



from .api import Base, query_cs
from .errors import ServerError
from .jobs import Jobs

# https://github.com/juju/juju/blob/master/api/facadeversions.go
_FACADE_VERSIONS = {
    "Action":                       0,
    "Addresser":                    1,
    "Agent":                        1,
    "AllWatcher":                   0,
    "AllEnvWatcher":                1,
    "Annotations":                  1,
    "Backups":                      0,
    "Block":                        1,
    "Charms":                       1,
    "CharmRevisionUpdater":         0,
    "Client":                       0,
    "Cleaner":                      1,
    "Deployer":                     0,
    "DiskManager":                  1,
    "EntityWatcher":                1,
    "Environment":                  0,
    "EnvironmentManager":           1,
    "FilesystemAttachmentsWatcher": 1,
    "Firewaller":                   1,
    "HighAvailability":             1,
    "ImageManager":                 1,
    "ImageMetadata":                1,
    "InstancePoller":               1,
    "KeyManager":                   0,
    "KeyUpdater":                   0,
    "LeadershipService":            1,
    "Logger":                       0,
    "MachineManager":               1,
    "Machiner":                     0,
    "MetricsManager":               0,
    "Networker":                    0,
    "NotifyWatcher":                0,
    "Pinger":                       0,
    "Provisioner":                  1,
    "Reboot":                       1,
    "RelationUnitsWatcher":         0,
    "Resumer":                      1,
    "Rsyslog":                      0,
    "Service":                      1,
    "Storage":                      1,
    "Spaces":                       1,
    "Subnets":                      1,
    "StorageProvisioner":           1,
    "StringsWatcher":               0,
    "SystemManager":                1,
    "Upgrader":                     0,
    "Uniter":                       2,
    "UserManager":                  0,
    "VolumeAttachmentsWatcher":     1,
}


class JujuClient(Base):
    API_VERSION = 1
    FACADE_VERSIONS = _FACADE_VERSIONS

    def status(self):
        """ Returns status of juju environment """
        return self.call(dict(Type="Client",
                              Request="FullStatus"),
                         timeout=60)

    def get_watcher(self):
        """ Returns watcher """
        return self.call(dict(Type="Client",
                              Request="WatchAll"))

    def get_watched_tasks(self, watcher_id):
        """ Returns a list of all watches for Id """
        return self.call(dict(Type="AllWatcher",
                              Request="Next",
                              Id=watcher_id))

    def add_charm(self, charm_url):
        """ Adds charm """
        return self.call(dict(Type="Client",
                              Request="AddCharm",
                              Params=dict(URL=charm_url)))

    def get_charm(self, charm_url):
        """ Get charm """
        return self.call(dict(Type='Client',
                              Request='CharmInfo',
                              Params=dict(CharmURL=charm_url)))

    def add_machine(self, series="", constraints={},
                    machine_spec="", parent_id="", container_type=""):
        """Allocate a new machine from the iaas provider.
        """
        if machine_spec:
            err_msg = "Cant specify machine spec with container_type/parent_id"
            assert not (parent_id or container_type), err_msg
            parent_id, container_type = machine_spec.split(":", 1)

        params = dict(
            Series=series,
            ContainerType=container_type,
            ParentId=parent_id,
            Constraints=self._prepare_constraints(constraints),
            Jobs=[Jobs.HostUnits])
        return self.add_machines([params])

    def add_machines(self, machines):
        """ Add machines """
        return self.call(dict(Type="Client",
                              Request="AddMachines",
                              Params=dict(MachineParams=machines)))

    def destroy_machines(self, machine_ids, force=False):
        params = {"MachineNames": machine_ids}
        if force:
            params["Force"] = True
        return self.call(dict(Type="Client",
                              Request="DestroyMachines",
                              Params=params))

    def resolved(self, unit_name, retry=0):
        """ Resolved """
        return self.call(dict(Type="Client",
                              Request="Resolved",
                              Params=dict(UnitName=unit_name,
                                          Retry=retry)))

    def get_public_address(self, target):
        """ Gets public address of instance """
        return self.call(dict(Type="Client",
                              Request="PublicAddress",
                              Params=dict(Target=target)))

    def info(self):
        """ Returns Juju environment state """
        return self.call(dict(Type="Client",
                              Request="EnvironmentInfo"))

    def get_env_constraints(self):
        """ Get environment constraints """
        return self.call(dict(Type="Client",
                              Request="GetEnvironmentConstraints"))

    def set_env_constraints(self, constraints):
        """ Set environment constraints """
        return self.call(dict(Type="Client",
                              Request="SetEnvironmentConstraints",
                              Params=constraints))

    def get_env_config(self):
        """ Get environment config """
        return self.call(dict(Type="Client",
                              Request="EnvironmentGet"))

    def set_env_config(self, config):
        """ Sets environment config variables """
        return self.call(dict(Type="Client",
                              Request="EnvironmentSet",
                              Params=dict(Config=config)))

    def add_relation(self, endpoint_a, endpoint_b):
        """ Adds relation between units """
        try:
            rv = self.call(dict(Type="Client",
                                Request="AddRelation",
                                Params=dict(Endpoints=[endpoint_a,
                                                       endpoint_b])))
        except ServerError as e:
            # do not treat pre-existing relations as exceptions:
            if 'relation already exists' in e.response['Error']:
                rv = e.response
            else:
                raise e

        return rv

    def remove_relation(self, endpoint_a, endpoint_b):
        """ Removes relation """
        return self.call(dict(Type="Client",
                              Request="DestroyRelaiton",
                              Params=dict(Endpoints=[endpoint_a,
                                                     endpoint_b])))

    def deploy(self, charm, service_name, num_units=1, config_yaml="",
               constraints=None, machine_spec=""):
        """ Deploy a charm to an instance

        :param str charm: Name of charm
        :param str service_name: name of service
        :param int num_units: number of units
        :param str config_yaml: charm configuration options
        :param dict constraints: deploy constraints
        :param str machine_spec: Type of machine to deploy to
        :returns: Deployed charm status
        """
        params = {'ServiceName': service_name}

        _url = query_cs(charm)
        params['CharmUrl'] = _url['charm']['url']
        params['NumUnits'] = num_units
        params['ConfigYAML'] = config_yaml

        if constraints:
            params['Constraints'] = self._prepare_constraints(
                constraints)
        if machine_spec:
            params['ToMachineSpec'] = machine_spec
        return self.call(dict(Type="Client",
                              Request="ServiceDeploy",
                              Params=dict(params)))

    def set_annotations(self, entity, entity_type, annotation):
        """ Sets annotations.
        :param dict annotation: dict with string pairs.
        """
        return self.call(dict(Type="Client",
                              Request="SetAnnotations",
                              Params=dict(Tag="%s-%s" % (entity_type, entity),
                                          Pairs=annotation)))

    def get_annotations(self, entity, entity_type):
        """ Gets annotations """
        return self.call(dict(Type="Client",
                              Request="GetAnnotations",
                              Params=dict(Tag="%s-%s" % (entity_type,
                                                         entity))))

    def set_config(self, service_name, config_keys):
        """ Sets machine config """
        return self.call(dict(Type="Client",
                              Request="ServiceSet",
                              Params=dict(ServiceName=service_name,
                                          Options=config_keys)))

    def unset_config(self, service_name, config_keys):
        """ Unsets machine config """
        return self.call(dict(Type="Client",
                              Request="ServiceUnset",
                              Params=dict(ServiceName=service_name,
                                          Options=config_keys)))

    def set_charm(self, service_name, charm_url, force=0):
        return self.call(dict(Type="Client",
                              Request="ServiceSetCharm",
                              Params=dict(ServiceName=service_name,
                                          CharmUrl=charm_url,
                                          Force=force)))

    def get_service(self, service_name):
        """ Get charm, config, constraits for srevice"""
        return self.call(dict(Type="Client",
                              Request="ServiceGet",
                              Params=dict(ServiceName=service_name)))

    def get_config(self, service_name):
        """ Get service configuration """
        svc = self.get_service(service_name)
        return svc['Config']

    def get_constraints(self, service_name):
        """ Get service constraints """
        return self.call(dict(Type="Client",
                              Request="GetServiceConstraints",
                              Params=dict(ServiceName=service_name)))

    def set_constraints(self, service_name, constraints):
        """ Sets service level constraints """
        return self.call(dict(Type="Client",
                              Request="SetServiceConstraints",
                              Params=dict(ServiceName=service_name,
                                          Constraints=constraints)))

    def update_service(self, service_name, charm_url, force_charm_url=0,
                       min_units=1, settings={}, constraints={}):
        """ Update service """
        return self.call(dict(Type="Client",
                              Request="SetServiceConstraints",
                              Params=dict(ServiceName=service_name,
                                          CharmUrl=charm_url,
                                          MinUnits=min_units,
                                          SettingsStrings=settings,
                                          Constraints=constraints)))

    def destroy_service(self, service_name):
        """ Destroy a service """
        return self.call(dict(Type="Client",
                              Request="ServiceDestroy",
                              Params=dict(ServiceName=service_name)))

    def expose(self, service_name):
        """ Expose a service """
        return self.call(dict(Type="Client",
                              Request="ServiceExpose",
                              Params=dict(ServiceName=service_name)))

    def unexpose(self, service_name):
        """ Unexpose service """
        return self.call(dict(Type="Client",
                              Request="ServiceUnexpose",
                              Params=dict(ServiceName=service_name)))

    def valid_relation_name(self, service_name):
        """ All possible relation names for service """
        return self.call(dict(Type="Client",
                              Request="ServiceCharmRelations",
                              Params=dict(ServiceName=service_name)))

    def add_unit(self, service_name, num_units=1, machine_spec=""):
        """ Add unit

        :param str service_name: Name of charm
        :param int num_units: Number of units
        :param str machine_spec: Type of machine to deploy to
        :returns dict: Units added
        """
        params = {}
        params['ServiceName'] = service_name
        params['NumUnits'] = num_units
        if machine_spec:
            params['ToMachineSpec'] = machine_spec

        return self.call(dict(Type="Client",
                              Request="AddServiceUnits",
                              Params=dict(params)))

    def remove_unit(self, unit_names):
        """ Removes unit """
        return self.call(dict(Type="Client",
                              Request="DestroyServiceUnits",
                              Params=dict(UnitNames=unit_names)))



# Copyright 2014-2016 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict, Counter
import copy
import logging
import yaml
from multiprocessing import cpu_count

from bundleplacer.maas import (satisfies, MaasMachineStatus)
from bundleplacer.state import ServiceState

from bundleplacer.assignmenttype import AssignmentType, label_to_atype
from bundleplacer.bundle import Bundle


log = logging.getLogger('bundleplacer')


DEFAULT_SHARED_ASSIGNMENT_TYPE = AssignmentType.LXD


class PlaceholderMachine:

    """A dummy MaasMachine that doesn't map to an actual machine in MAAS.

    To specify a future virtual machine for placement that will be
    created later, pass in vm specs as juju constraints in 'constraints'.

    The keys juju uses differ somewhat from the MAAS API status keys,
    and are mapped so that they will appear correct to placement code
    expecting MAAS machines.
    """

    def __init__(self, instance_id, name, constraints=None):
        self.instance_id = instance_id
        self.system_id = instance_id
        self.machine_id = -1
        self.display_name = name
        def_cons = {'arch': '?',
                    'cpu_count': 0,
                    'cpu_cores': 0,
                    'memory': 0,
                    'mem': 0,
                    'storage': 0}
        if constraints is None:
            self.constraints = def_cons
        else:
            self.constraints = constraints

    @property
    def arch(self):
        return self.constraints['arch']

    @property
    def cpu_cores(self):
        return self.constraints['cpu_cores']

    def filter_label(self):
        return self.display_name

    @property
    def machine(self):
        return self.constraints

    @property
    def mem(self):
        return self.constraints['mem']

    @property
    def status(self):
        return MaasMachineStatus.UNKNOWN

    @property
    def storage(self):
        return self.constraints['storage']

    @property
    def hostname(self):
        return self.display_name

    def __repr__(self):
        return "<Placeholder Machine: {}>".format(self.display_name)


class PlacementError(Exception):

    "Generic exception class for placement related errors"


class PlacementController:

    """Keeps state of current machines and their assigned services.

    Maintains two placeholder machines, one for subordinate charms and
    one for "Juju Default" that are both the equivalent of not
    specifying a machine to deploy to when invoking Juju.

    """

    def __init__(self, maas_state=None, config=None):
        self.config = config
        self.maas_state = maas_state
        self.maasinfo = defaultdict(lambda: '?')
        if self.maas_state:
            sn = self.maas_state.get_server_config('maas_name')
            self.maasinfo['server_name'] = sn
            self.maasinfo['server_hostname'] = self.maas_state.server_hostname
        self._machines = []
        self._bundle_placeholders = []
        self.sub_placeholder = PlaceholderMachine('_subordinates',
                                                  'Subordinate Charms')
        self.def_placeholder = PlaceholderMachine('_default',
                                                  'Juju Default')
        # assignments is {id: {atype: [service]}}
        self.assignments = defaultdict(lambda: defaultdict(list))
        self.deployments = defaultdict(lambda: defaultdict(list))
        self.autosave_filename = None
        mf = config.getopt('metadata_filename')
        self.bundle = Bundle(filename=config.getopt('bundle_filename'),
                             metadatafilename=mf)
        if self.config.getopt('provider_type') == "lxd":
            self.bundle.clear_machines_and_placement()

        self.update_from_bundle()
        self.reset_assigned_deployed()

    def get_temp_copy(self):
        """Returns another PlacementController that can be used to track
        assignments temporarily, e.g. for supporting cancellable
        assignments in a dialog box.

        Pairs with update_from_controller() to 'commit' those temporary
        assignments to the 'main' controller.
        """
        newpc = PlacementController(maas_state=self.maas_state,
                                    config=self.config)
        newpc.assignments = copy.copy(self.assignments)
        newpc.deployments = copy.copy(self.deployments)
        newpc._machines = self._machines
        newpc.reset_assigned_deployed()
        return newpc

    def update_from_controller(self, other):
        """Updates internal structures based on other's.
        For integrating temporarily tracked updates."""

        self.assignments = other.assignments
        self.deployments = other.deployments
        self.reset_assigned_deployed()

    def set_assignments_from_deployments(self):
        """Reset deployment state of all services. Useful after reading a file
        from a previous install.
        """
        self.assignments = self.deployments
        self.deployments = defaultdict(lambda: defaultdict(list))
        self.reset_assigned_deployed()

    def __repr__(self):
        return "<PlacementController {}>".format(id(self))

    def set_autosave_filename(self, filename):
        self.autosave_filename = filename

    def do_autosave(self):
        if not self.autosave_filename:
            return
        with open(self.autosave_filename, 'w') as af:
            self.save(af)

    def save(self, f):
        """f is a file-like object to save state to, to be re-read by
        load(). No guarantees made about the contents of the file.
        """
        flat_assignments = defaultdict(dict)
        for iid, ad in self.assignments.items():

            flat_ad = {}
            for atype, al in ad.items():
                flat_al = [cc.service_name for cc in al]
                flat_ad[atype.name] = flat_al

            flat_assignments[iid]['assignments'] = flat_ad

        for iid, dd in self.deployments.items():
            flat_dd = {}
            for atype, dl in dd.items():
                flat_dl = [cc.service_name for cc in dl]
                flat_dd[atype.name] = flat_dl
            flat_assignments[iid]['deployments'] = flat_dd

        for iid in flat_assignments.keys():
            constraints = {}
            if self.maas_state is None:
                machine = next((m for m in self.machines() if
                                m.instance_id == iid), None)
                if machine:
                    constraints = machine.constraints
                    flat_assignments[iid]['constraints'] = constraints

        yaml.dump(flat_assignments, f)

    def load(self, f):
        """Load assignments from file object written to by save().
        replaces current assignments.
        """
        def find_service(name):
            for s in self.services():
                if s.service_name == name:
                    return s
            log.warning("Could not find service "
                        "matching saved service name {}".format(name))
            return None

        file_assignments = yaml.load(f)
        new_assignments = defaultdict(lambda: defaultdict(list))
        new_deployments = defaultdict(lambda: defaultdict(list))
        for iid, d in file_assignments.items():
            if self.maas_state is None and \
               not self.is_placeholder(iid):
                constraints = d.get('constraints', {})
                pm = PlaceholderMachine(iid, iid,
                                        constraints)
                self._machines.append(pm)

            ad = d.get('assignments', {})
            for atypestr, al in ad.items():
                new_al = [find_service(ccname)
                          for ccname in al]
                new_al = [x for x in new_al if x is not None]
                at = AssignmentType.__members__[atypestr]
                new_assignments[iid][at] = new_al

            dd = d.get('deployments', {})
            for atypestr, dl in dd.items():
                new_dl = [find_service(ccname)
                          for ccname in dl]
                new_dl = [x for x in new_dl if x is not None]
                at = AssignmentType.__members__[atypestr]
                new_deployments[iid][at] = new_dl

        self.assignments.clear()
        self.assignments.update(new_assignments)
        self.deployments.clear()
        self.deployments.update(new_deployments)
        self.reset_assigned_deployed()

    def update_and_save(self):
        self.reset_assigned_deployed()
        self.do_autosave()

    def is_placeholder(self, mid):
        return mid in [self.sub_placeholder.instance_id,
                       self.def_placeholder.instance_id]

    def machines(self, include_placeholders=True):
        """Returns all machines known to the controller.

        if 'include_placeholder' is False, any placeholder machines
        are excluded.
        """
        if self.maas_state:
            cons = self.config.getopt('constraints')
            ms = self.maas_state.machines(constraints=cons)
        else:
            ms = self._machines

        if include_placeholders:
            return ms + [self.sub_placeholder, self.def_placeholder] + \
                self._bundle_placeholders
        else:
            return ms

    def machines_pending(self, include_placeholders=False):
        """Returns a list of machines that have services assigned to them
        which are not yet deployed.

        Excludes placeholder machines by default, so this can be used
        to e.g. get the number of real machines to wait for.

        """
        ms = []
        for m in self.machines(include_placeholders=include_placeholders):
            if m.instance_id in self.assignments:
                n = sum(len(cl) for _, cl in
                        self.assignments[m.instance_id].items())
                if n > 0:
                    ms.append(m)
        return ms

    def add_new_service(self, charm_name, charm_dict, service_name=None):
        """adds a service with the default name of 'charm_name' or
        'charm_name-1', etc"""
        self.bundle.add_new_service(charm_name, charm_dict, service_name)

    def update_from_bundle(self):
        self.add_bundle_machines(self.bundle.machines)
        self.add_bundle_assignments(self.bundle.assignments)
        self.add_subordinates(self.bundle.services)

    def merge_bundle(self, bundle_dict):
        new_bundle = Bundle(bundle_data=bundle_dict)
        if self.config.getopt('provider_type') == "lxd":
            new_bundle.clear_machines_and_placement()

        t = self.bundle.update(new_bundle)
        new_machines, new_services, new_assignments = t
        self.add_bundle_machines(new_machines)
        self.add_bundle_assignments(new_assignments)
        self.add_subordinates(new_services)
        return new_bundle

    def add_bundle_machines(self, machines):
        for mid, md in machines.items():
            pm = PlaceholderMachine(mid, "bundle-machine-" + mid,
                                    md.get('constraints', {}))
            self._bundle_placeholders.append(pm)

    def add_bundle_assignments(self, new_as):
        for sname, tostrs in new_as.items():
            service = next((s for s in self.bundle.services
                            if s.service_name == sname), None)
            if service is None:
                continue
            for tostr in tostrs:
                parts = tostr.split(":")
                if len(parts) > 1:
                    atype, mid = parts
                    atype = label_to_atype([atype])[0]
                else:
                    atype, mid = AssignmentType.DEFAULT, parts[0]
                machine = next((m for m in self.machines() if
                                m.instance_id == mid), None)
                if machine:
                    self.assign(machine, service, atype)

    def add_subordinates(self, all_services):
        """looks through all_services and assigns any subordinates to the
        subordinate placeholder."""
        d = self.assignments[self.sub_placeholder.instance_id]
        al = d[AssignmentType.DEFAULT]
        al += [s for s in all_services if s.subordinate]

    def remove_service(self, service_name):
        self.bundle.remove_service(service_name)

    def toggle_relation(self, s1_name, s1_rel, s2_name, s2_rel):
        if self.bundle.is_related(s1_name, s1_rel, s2_name, s2_rel):
            self.bundle.remove_relation(s1_name, s1_rel, s2_name, s2_rel)
        else:
            self.bundle.add_relation(s1_name, s1_rel, s2_name, s2_rel)

    def is_related(self, s1_name, s1_rel, s2_name, s2_rel):
        return self.bundle.is_related(s1_name, s1_rel, s2_name, s2_rel)

    def services(self):
        return self.bundle.services

    def charm_names(self):
        seen = set()
        return [s.charm_name for s in self.bundle.services
                if s.charm_name not in seen and not seen.add(s.charm_name)]

    def services_with_charm(self, charm_name):
        return [s for s in self.bundle.services
                if s.charm_name == charm_name]

    @property
    def assigned_services(self):
        """Returns a deduplicated list of all services that have a placement
        assigned, but are not yet deployed.

        """
        return list(self._assigned_services)

    @property
    def deployed_services(self):
        """Returns a deduplicated list of all services that have been deployed.
        """
        return list(self._deployed_services)

    def assign(self, machine, service, atype):
        if not service.allow_multi_units:
            for m, d in self.assignments.items():
                for at, l in d.items():
                    if service in l:
                        l.remove(service)

        self.assignments[machine.instance_id][atype].append(service)
        self.update_and_save()

    def mark_deployed(self, machine, service, atype):
        self.deployments[machine.instance_id][atype].append(service)
        self.assignments[machine.instance_id][atype].remove(service)
        self.update_and_save()

    def _get_machines_by_atype(self, a_dict, service):
        "Helper for get_assignments and get_deployments"
        all_machines = self.machines()

        machines_by_atype = defaultdict(list)
        for m_id, d in a_dict.items():
            m = next((m for m in all_machines
                      if m.instance_id == m_id), None)
            if not m:
                log.debug("can't find machine for m_id '{}'".format(m_id))
                continue

            for atype, assignment_list in d.items():
                for c in assignment_list:
                    if c == service:
                        machines_by_atype[atype].append(m)

        return machines_by_atype

    def get_assignments(self, service):
        """returns assignments for a given service

        returns a dict like {assignment_type : [machines]}
        """
        return self._get_machines_by_atype(self.assignments,
                                           service)

    def get_deployments(self, service):
        """returns deployments for a given service

        returns a dict like {assignment_type : [machines]}
        """
        return self._get_machines_by_atype(self.deployments,
                                           service)

    def clear_all_assignments(self):
        self.assignments = defaultdict(lambda: defaultdict(list))
        self.update_and_save()

    def clear_assignments(self, m):
        """clears all assignments for machine m.
        If m has no assignments, does nothing.
        """
        if m.instance_id not in self.assignments:
            return

        del self.assignments[m.instance_id]
        self.update_and_save()

    def remove_one_assignment(self, m, cc):
        ad = self.assignments[m.instance_id]
        for atype, assignment_list in ad.items():
            if cc in assignment_list:
                assignment_list.remove(cc)
                break
        self.update_and_save()

    def assignments_for_machine(self, m):
        """Returns all assignments for given machine

        {assignment_type: [service]}
        """
        return self.assignments[m.instance_id]

    def deployments_for_machine(self, m):
        """Returns deployments
        {atype: [service]}
        """
        return self.deployments[m.instance_id]

    def is_assigned_to(self, service, machine):
        assignment_dict = self.assignments[machine.instance_id]
        for atype, services in assignment_dict.items():
            if service in services:
                return True
        return False

    def is_deployed_to(self, service, machine):
        dd = self.deployments[machine.instance_id]
        for atype, services in dd.items():
            if service in services:
                return True
        return False

    def set_all_assignments(self, assignments):
        self.assignments = assignments
        self.update_and_save()

    def reset_assigned_deployed(self):
        self._assigned_services = set()
        self._deployed_services = set()
        for cc in self.services():
            ad = self.get_assignments(cc)
            is_assigned = False
            for atype, al in ad.items():
                if len(al) > 0:
                    is_assigned = True
            if is_assigned:
                self._assigned_services.add(cc)

            dd = self.get_deployments(cc)
            is_deployed = False
            for atype, dl in dd.items():
                if len(dl) > 0:
                    is_deployed = True
            if is_deployed:
                self._deployed_services.add(cc)

    def is_assigned(self, service):
        return service in self._assigned_services

    def is_deployed(self, service):
        return service in self._deployed_services

    def get_service_state(self, service):
        """Returns tuple of service state:
        (state, cons, deps)

        state is a ServiceState:

        - REQUIRED means that the service still must be assigned before
        deploying is OK.

        IF a service dependency forced this, then the other service will
        be in 'deps'.  'deps' is NOT just a list of all services that
        depend on the given service.

        - CONFLICTED means that it can't be assigned until a conflicting
        service is unassigned.  In this case, the conflicting service is in
        'cons'.

        - OPTIONAL means that it is ok either way. deps and cons are unused

        """
        state = ServiceState.OPTIONAL
        conflicting = set()
        depending = set()

        def conflicts_with(other_service):
            return (service.service_name in other_service.conflicts or
                    other_service.service_name in service.conflicts)

        def depends(a_service, b_service):
            return b_service.service_name in a_service.depends

        required_services = [c for c in self.services()
                             if c.is_core]

        planned_or_deployed = (self.assigned_services +
                               required_services +
                               self.deployed_services)

        for other_service in planned_or_deployed:
            if conflicts_with(other_service):
                state = ServiceState.CONFLICTED
                conflicting.add(other_service)
            if depends(other_service, service):
                if state != ServiceState.CONFLICTED:
                    state = ServiceState.REQUIRED
                depending.add(other_service)

        if service.service_name in [c.service_name for c in required_services]:
            state = ServiceState.REQUIRED

        n_required = service.required_num_units()
        # sanity check:
        if n_required > 1 and not service.allow_multi_units:
            log.error("Inconsistent service definition for {}:"
                      " - requires {} units but does not allow "
                      "multi units.".format(service.service_name, n_required))

        n_units = (self.assignment_machine_count_for_service(service) +
                   self.deployment_machine_count_for_service(service))

        if state == ServiceState.OPTIONAL and \
           n_units > 0 and n_units < n_required:
            state = ServiceState.REQUIRED
        elif state == ServiceState.REQUIRED and n_units >= n_required:
            if n_units > 0:
                state = ServiceState.OPTIONAL

        return (state, list(conflicting), list(depending))

    def unassigned_undeployed_services(self):
        all_services = set(self.services())
        return (all_services -
                (self._assigned_services.union(self._deployed_services)))

    def can_deploy(self):
        unassigned_requireds = [cc for cc in
                                self.unassigned_undeployed_services()
                                if self.get_service_state(cc)[0] ==
                                ServiceState.REQUIRED]

        underassigned = [cc for cc in self._assigned_services if
                         (self.assignment_machine_count_for_service(cc) +
                          self.deployment_machine_count_for_service(cc)) <
                         cc.required_num_units()]
        return len(unassigned_requireds) + len(underassigned) == 0

    def assignment_machine_count_for_service(self, cc):
        """Returns the total number of assignments of any type for a given
        service."""
        return sum([len(al) for al in self.get_assignments(cc).values()])

    def deployment_machine_count_for_service(self, cc):
        """Returns the total number of deployments of any type for a given
        service."""
        return sum([len(al) for al in self.get_deployments(cc).values()])

    def autoassign_unassigned_services(self):
        """Attempt to find machines for all required unassigned services using
        only empty machines.

        Returns a pair (success, message) where success is True if all
        services are assigned. message is an info message for the user.

        """

        empty_machines = [m for m in self.machines(include_placeholders=False)
                          if len(self.assignments[m.instance_id]) == 0]

        unassigned_services = list(self.unassigned_undeployed_services())
        unassigned_defaults = self.gen_defaults(unassigned_services,
                                                empty_machines)

        for mid, services in unassigned_defaults.items():
            self.assignments[mid] = services

        self.update_and_save()

        unassigned_services = list(self.unassigned_undeployed_services())
        unassigned_reqs = [c for c in unassigned_services if
                           self.get_service_state(c)[0] ==
                           ServiceState.REQUIRED]

        if len(unassigned_reqs) > 0:
            msg = ("Not enough empty machines could be found for the "
                   "following required services. Please add machines and "
                   "try again, or finish placement manually.")
            m = ", ".join([c.service_name for c in unassigned_reqs])
            return (False, msg + "\n" + m)
        return (True, "")

    def autoassign_unassigned_to_default(self):
        """Assigns all unassigned *non-subordinate* services to juju default
        placeholder.
        """
        for s in self.unassigned_undeployed_services():
            d = self.assignments[self.def_placeholder.instance_id]
            al = d[DEFAULT_SHARED_ASSIGNMENT_TYPE]
            for i in range(s.num_units):
                al.append(s)
        self.update_and_save()

    def gen_defaults(self, services=None, maas_machines=None):
        """Generates an assignments dictionary for the given service classes and
        machines, based on constraints.

        Does not alter controller state.

        Use set_all_assignments(gen_defaults()) to clear and reset the
        controller's state to these defaults.

        Should not be used for single installs, see gen_single.
        """
        if self.maas_state is None:
            raise PlacementError("Can't call gen_defaults with no maas_state")

        if services is None:
            services = self.services()
        log.debug("in gen_defaults, services is {}".format(services))
        assignments = defaultdict(lambda: defaultdict(list))

        if maas_machines is None:
            maas_machines = self.maas_state.machines(
                MaasMachineStatus.READY,
                constraints=self.config.getopt('constraints'))

        def satisfying_machine(constraints):
            for machine in maas_machines:
                if satisfies(machine, constraints)[0]:
                    maas_machines.remove(machine)
                    return machine

            return None

        isolated_services, controller_services = [], []
        subordinate_services = []

        for service in services:
            state, _, _ = self.get_service_state(service)
            if state != ServiceState.REQUIRED:
                continue
            if service.isolate:
                assert(not service.subordinate)
                isolated_services.append(service)
            elif service.subordinate:
                assert(not service.isolate)
                subordinate_services.append(service)
            else:
                controller_services.append(service)

        for service in isolated_services:
            for n in range(service.required_num_units()):
                m = satisfying_machine(service.constraints)
                if m:
                    l = assignments[m.instance_id][AssignmentType.BareMetal]
                    l.append(service)

        controller_machine = satisfying_machine({})
        if controller_machine:
            for service in controller_services:
                ad = assignments[controller_machine.instance_id]
                l = ad[DEFAULT_SHARED_ASSIGNMENT_TYPE]
                l.append(service)

        for service in subordinate_services:
            ad = assignments[self.sub_placeholder.instance_id]
            l = ad[AssignmentType.DEFAULT]
            l.append(service)

        import pprint
        log.debug(pprint.pformat(assignments))
        return assignments

    def gen_single(self):
        """Generates an assignment for the single installer."""
        assignments = defaultdict(lambda: defaultdict(list))

        max_cpus = cpu_count()
        if max_cpus >= 2:
            max_cpus = min(8, max_cpus // 2)

        controller = PlaceholderMachine('controller', 'controller',
                                        {'mem': 6144,
                                         'root-disk': 20480,
                                         'cpu-cores': max_cpus})
        self._machines.append(controller)

        service_name_counter = Counter()

        def placeholder_for_service(service):
            mnum = service_name_counter[service.service_name]
            service_name_counter[service.service_name] += 1

            instance_id = '{}-machine-{}'.format(service.service_name,
                                                 mnum)
            m_name = 'machine {} for {}'.format(mnum,
                                                service.display_name)

            return PlaceholderMachine(instance_id, m_name,
                                      service.constraints)

        for service in self.services():
            state, _, _ = self.get_service_state(service)
            if state != ServiceState.REQUIRED:
                continue
            if service.isolate:
                assert(not service.subordinate)
                for n in range(service.required_num_units()):
                    pm = placeholder_for_service(service)
                    self._machines.append(pm)
                    ad = assignments[pm.instance_id]
                    ad[AssignmentType.DEFAULT].append(service)
            elif service.subordinate:
                assert(not service.isolate)
                ad = assignments[self.sub_placeholder.instance_id]
                l = ad[AssignmentType.DEFAULT]
                l.append(service)
            else:
                ad = assignments[controller.instance_id]
                ad[AssignmentType.LXD].append(service)

        import pprint
        log.debug("gen_single() = '{}'".format(pprint.pformat(assignments)))
        return assignments


class BundleWriter:
    def __init__(self, controller):
        self.controller = controller

    def _dict_for_service(self, svc, atype, to, services):

        tolist = []
        num_units = 1
        if svc.service_name in services:
            num_units = services[svc.service_name]['num_units'] + 1
            tolist = services[svc.service_name].get('to', [])

        d = dict(charm=svc.charm_source,
                 options=svc.options)
        if not svc.subordinate:
            d['num_units'] = num_units

        if to is not None:
            prefix = {AssignmentType.DEFAULT: "",
                      AssignmentType.BareMetal: "",
                      AssignmentType.KVM: "kvm:",
                      AssignmentType.LXD: "lxd:",
                      AssignmentType.LXC: "lxc:"}[atype]
            tolist.append("{}{}".format(prefix, to))

        if len(tolist) > 0:
            d['to'] = tolist
        return d

    def _dict_for_machine(self, mid):
        machine_tag = mid.split('/')[-2]
        cstr = "tags={}".format(machine_tag)
        return {"constraints": cstr}

    def _get_used_relations(self, services):
        relations = []
        service_names = [s.service_name for s in services]
        for svc in services:
            for src, dst in svc.relations:
                src_service = src.split(":")[0]
                dst_service = dst.split(":")[0]
                if src_service in service_names and \
                   dst_service in service_names:
                    relations.append([src, dst])
        # uniquify list of relations
        seen = set()
        return [r for r in relations
                if str(r) not in seen and not seen.add(str(r))]

    def write_bundle(self, filename):
        bundle = {}
        services = {}
        servicenames = []
        machines = self.controller.bundle.machines
        iid_map = {}            # maps iid to juju machine number

        self.controller.autoassign_unassigned_to_default()

        # get a machine dict for every machine with at least one
        # service
        existing_ids = list(machines.keys()) + ["_subordinates", "_default"]
        for mid in machines.keys():
            iid_map[mid] = mid

        for iid, d in self.controller.assignments.items():
            if sum([len(svcs) for svcs in d.values()]) == 0:
                continue

            if iid not in existing_ids:
                machine_id = len(machines) + 1
                iid_map[iid] = machine_id
                machines[machine_id] = self._dict_for_machine(iid)

        for iid, d in self.controller.assignments.items():
            for atype, svcs in d.items():
                if len(svcs) < 1:
                    continue
                for svc in svcs:
                    sd = self._dict_for_service(svc, atype,
                                                iid_map.get(iid, None),
                                                services)
                    services[svc.service_name] = sd
                    servicenames.append(svc)

        bundle['machines'] = machines
        bundle['services'] = services
        bundle['relations'] = self._get_used_relations(servicenames)

        for k, v in self.controller.bundle.extra_items().items():
            bundle[k] = v

        with open(filename, 'w') as f:
            yaml.dump(bundle, f, default_flow_style=False)



# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from urwid import Divider, Pile, Text, WidgetWrap

from bundleplacer.ui.simple_service_widget import SimpleServiceWidget

log = logging.getLogger('bundleplacer.ui')


class ServicesList(WidgetWrap):

    """A list of services with flexible display options.

    Note that not all combinations of display options make sense. YMMV.

    controller - a PlacementController

    display_controller - a PlacerView

    machine - a machine instance to query for constraint checking. If
    None, no constraint checking is done. If set, only services whose
    constraints are satisfied by 'machine' are shown.

    """

    def __init__(self, placement_controller, display_controller,
                 show_placements=False, title="Services"):
        self.placement_controller = placement_controller
        self.display_controller = display_controller
        self.service_widgets = []
        self.show_placements = show_placements
        self.title = title
        w = self.build_widgets()
        self.update()
        super().__init__(w)

    def selectable(self):
        # overridden to ensure that we can arrow through the buttons
        # shouldn't be necessary according to documented behavior of
        # Pile & Columns, but discovered via trial & error.
        return True

    def build_widgets(self):
        widgets = []
        if self.title:
            widgets = [Text(self.title), Divider(' ')]
        widgets += self.service_widgets
        self.service_pile = Pile(widgets)
        return self.service_pile

    def focus_top_or_next(self):
        return self._setfocus(top=False)

    def focus_top(self):
        return self._setfocus(top=True)

    def _setfocus(self, top):
        try:
            if top:
                self.service_pile.focus_position = 0
            else:
                pos = self.service_pile.focus_position
                if pos + 1 >= len(self.service_widgets):
                    return False
                self.service_pile.focus_position = pos + 1
        except IndexError:
            log.debug("caught indexerror in servicescolumn.focus_next")
            return False
        return True

    def focused_service_widget(self):
        if len(self.service_widgets) > 0:
            return self.service_pile.focus
        return None

    def find_service_widget(self, s):
        return next((sw for sw in self.service_widgets if
                     sw.service.service_name == s.service_name),
                    None)

    def update(self):
        for s in self.placement_controller.services():

            sw = self.find_service_widget(s)
            if sw is None:
                sw = self.add_service_widget(s)
            sw.update()

        allnames = [s.service_name
                    for s in self.placement_controller.services()]
        for sw in self.service_widgets:
            if sw.service.service_name not in allnames:
                self.remove_service_widget(sw.service)

        self.sort_service_widgets()

    def add_service_widget(self, service):
        sw = SimpleServiceWidget(service, self.placement_controller,
                                 self.display_controller,
                                 show_placements=self.show_placements)
        self.service_widgets.append(sw)
        options = self.service_pile.options()
        self.service_pile.contents.append((sw, options))

        # NOTE: see the + 1: indexing in remove_service_widget if you
        # re-add this divider. it should then be +2.

        # self.service_pile.contents.append((AttrMap(Padding(Divider('\u23bc'),
        #                                                    left=2, right=2),
        #                                            'label'), options))
        return sw

    def remove_service_widget(self, service):
        sw = self.find_service_widget(service)

        if sw is None:
            return
        self.service_widgets.remove(sw)
        sw_idx = 0
        for w, opts in self.service_pile.contents:
            if w == sw:
                break
            sw_idx += 1

        c = self.service_pile.contents[:sw_idx] + \
            self.service_pile.contents[sw_idx + 1:]
        self.service_pile.contents = c

    def sort_service_widgets(self):
        def keyfunc(sw):
            s = sw.service
            if s.subordinate:
                skey = 'z'
            else:
                skey = s.service_name
            return skey
        self.service_widgets.sort(key=keyfunc)

        def wrappedkeyfunc(t):
            mw, options = t
            if not isinstance(mw, SimpleServiceWidget):
                return 'A'
            return keyfunc(mw)

        self.service_pile.contents.sort(key=wrappedkeyfunc)

    def select_service(self, service_name):
        idx = 0
        for w, opts in self.service_pile.contents:
            if w.service.service_name == service_name:
                self.service_pile.focus_position = idx
                return
            idx += 1



from ubuntui.widgets.buttons import (confirm_btn, cancel_btn)
from ubuntui.widgets.text import Instruction
from ubuntui.utils import Color, Padding
from ubuntui.widgets.hr import HR
from urwid import (WidgetWrap, Pile, Text, Columns, Filler)


class NewCloudView(WidgetWrap):
    def __init__(self, app, cloud, schema, cb):
        self.app = app
        self.cloud = cloud
        self.input_items = schema
        self.cb = cb
        _pile = [
            Padding.center_60(Instruction(
                "Enter your {} credentials:".format(self.cloud.upper()))),
            Padding.center_60(HR()),
            Padding.center_60(self.build_inputs()),
            Padding.line_break(""),
            Padding.center_20(self.buttons())
        ]
        super().__init__(Filler(Pile(_pile), valign="middle"))

    def _swap_focus(self):
        if self._w.body.focus_position == 2:
            self._w.body.focus_position = 4
        else:
            self._w.body.focus_position = 2

    def keypress(self, size, key):
        if key in ['tab', 'shift tab']:
            self._swap_focus()
        return super().keypress(size, key)

    def buttons(self):
        confirm = confirm_btn(on_press=self.submit)
        cancel = cancel_btn(on_press=self.cancel)

        buttons = [
            Color.button_primary(confirm, focus_map='button_primary focus'),
            Color.button_secondary(cancel, focus_map='button_secondary focus')
        ]
        return Pile(buttons)

    def build_inputs(self):
        items = []
        for k in self.input_items.keys():
            display = k
            if k.startswith('_'):
                # Don't treat 'private' keys as input
                continue
            if k.startswith('@'):
                # Strip public, not storable attribute
                display = k[1:]
            col = Columns(
                [
                    ('weight', 0.5, Text(display, align='right')),
                    Color.string_input(self.input_items[k],
                                       focus_map='string_input focus')
                ], dividechars=1
            )
            items.append(col)
            items.append(Padding.line_break(""))
        return Pile(items)

    def cancel(self, btn):
        self.cb(back=True)

    def submit(self, result):
        self.cb(self.input_items)



class BundleModelException(Exception):
    """ Exception in BundleModel """


class BundleModel:
    """ Stores bundle location for juju deploy
    """
    bundle = {
        "key": None,
        "name": None,
        "summary": None,
        "revision": None,
        "location": None,
        "blacklist": [],
        "whitelist": [],
        "recommendedCharms": [],
        "bootstrapSeries": None
    }

    @classmethod
    def key(cls):
        """ Returns key of resource
        """
        return cls.bundle.get('key', None)

    @classmethod
    def location(cls):
        """ Location of bundle, this will override key as
        location can contain namespaced bundles
        """
        return cls.bundle.get('location', None)

    @classmethod
    def bootstrapSeries(cls):
        """ Returns a defined bootstrap-series
        """
        return cls.bundle.get('bootstrapSeries', None)

    @classmethod
    def name(cls):
        """ Returns name of resource
        """
        return cls.bundle.get('name', None)

    @classmethod
    def summary(cls):
        """ Returns summary of resource
        """
        return cls.bundle.get('summary', None)

    @classmethod
    def revision(cls):
        """ Returns revision of resource
        """
        return cls.bundle.get('revision', None)

    @classmethod
    def whitelist(cls):
        """ Returns whitelisted provider types
        """
        return cls.bundle.get('whitelist', [])

    @classmethod
    def blacklist(cls):
        """ Returns blacklisted provider types
        """
        return cls.bundle.get('blacklist', [])

    @classmethod
    def recommended(cls):
        """ Returns recommended charms
        """
        return cls.bundle.get('recommendedCharms', [])

    @classmethod
    def to_entity(cls, use_latest=True):
        """ Returns proper entity key to query the charmstore.

        Arguments:
        use_latest: Use latest bundle revision from charmstore

        Returns:
        Formatted entity string suitable for charmstore lookup
        """
        if cls.location() is None and cls.key() is None:
            raise BundleModelException("Unable to determine bundle path.")
        if cls.location():
            return cls.location()
        if not use_latest:
            return "{}-{}".format(cls.key(), cls.revision())
        return cls.key()

    @classmethod
    def to_path(cls):
        """ Returns bundle path suitable for juju deploy <bundle>
        """
        bundle = cls.key()
        if bundle is None:
            raise BundleModelException("Unable to determine bundle")
        if cls.revision() is not None:
            bundle = "{}-{}".format(bundle, cls.revision())
        return "cs:bundle/{}".format(bundle)



# Copyright 2014-2016 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from urwid import Divider, Pile, WidgetWrap

from bundleplacer.ui.services_list import ServicesList
from bundleplacer.ui.simple_service_widget import ServiceWidgetState


log = logging.getLogger('bundleplacer')


class ServicesColumn(WidgetWrap):

    """Displays dynamic list of unplaced services and associated controls
    """

    def __init__(self, display_controller, placement_controller,
                 placement_view):
        self.display_controller = display_controller
        self.placement_controller = placement_controller
        self.placement_view = placement_view
        w = self.build_widgets()
        super().__init__(w)
        self.update()

    def selectable(self):
        return True

    def build_widgets(self):
        self.services_list = ServicesList(self.placement_controller,
                                          self.display_controller,
                                          title=None)

        self.services_pile = Pile([self.services_list, Divider()])

        return self.services_pile

    def focus_top(self):
        self.update()
        self.services_list.focus_top()

    def focus_next(self):
        self.update()
        moved = self.services_list.focus_top_or_next()
        fsw = self.services_list.focused_service_widget()
        if not moved or (fsw and fsw.service.subordinate):
            self.placement_view.focus_footer()

    def update(self):
        self.services_list.update()

    def do_reset_to_defaults(self, sender):
        self.placement_controller.set_all_assignments(
            self.placement_controller.gen_defaults())

    def clear_selections(self):
        for sw in self.services_list.service_widgets:
            sw.state = ServiceWidgetState.UNSELECTED

    def select_service(self, service_name):
        self.services_list.select_service(service_name)



#
# machine.py - Machine
#
# Copyright 2014 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging


log = logging.getLogger('bundleplacer')


class Machine:

    """ Base machine class """

    def __init__(self, machine_id, machine):
        self.machine_id = machine_id
        self.machine = machine
        self._cpu_cores = self.hardware('cpu-cores')
        self._storage = self.hardware('root-disk')
        self._mem = self.hardware('memory')
        self.agent = self.machine.get('Agent', None)
        self.agent_state = self.machine.get('AgentState', None)
        self.agent_state_info = self.machine.get('AgentStateInfo', None)
        self.agent_version = self.machine.get('AgentVersion', None)
        self.dns_name = self.machine.get('DNSName', '')
        self.err = self.machine.get('Err', None)
        self.has_vote = self.machine.get('HasVote')
        self.wants_vote = self.machine.get('WantsVote')

    @property
    def instance_id(self):
        """ Returns InstanceId

        :returns: instance ID
        :rtype: str
        """
        return self.machine.get('InstanceId', None)

    @property
    def cpu_cores(self):
        """ Return number of cpu-cores

        :returns: number of cpus
        :rtype: str
        """
        return self._cpu_cores

    @cpu_cores.setter
    def cpu_cores(self, val):
        self._cpu_cores = val

    @property
    def arch(self):
        """ Return architecture

        :returns: architecture type
        :rtype: str
        """
        return self.hardware('arch')

    @property
    def storage(self):
        """ Return storage

        :returns: storage size
        :rtype: str
        """
        try:
            self._storage = int(self._storage[:-1]) / 1024
            return "{size}G".format(size=str(self._storage))
        except:
            return "N/A"

    @storage.setter
    def storage(self, val):
        self._storage = val

    @property
    def mem(self):
        """ Return memory

        :returns: memory size
        :rtype: str
        """
        return "{size}".format(size=str(self._mem))

    @mem.setter
    def mem(self, val):
        self._mem = val

    def hardware(self, spec):
        """ Get hardware information

        :param spec: a hardware specification
        :type spec: str
        :returns: hardware of spec
        :rtype: str
        """
        _machine = self.machine.get('Hardware', None)
        if _machine:
            _hardware_list = _machine.split(' ')
            for item in _hardware_list:
                k, v = item.split('=')
                if k in spec:
                    return v
        return "N/A"

    @property
    def containers(self):
        """ Return containers for machine

        :rtype: generator
        """
        _containers = self.machine.get('Containers', {}).items()
        for container_id, container in _containers:
            yield Machine(container_id, container)

    def container(self, container_id):
        """ Inspect a container

        :param container_id: lxd container id
        :type container_id: int
        :returns: Returns a dictionary of the container information for
                  specific machine and lxd id.
        :rtype: dict
        """
        for m in self.containers:
            if m.machine_id == container_id:
                return m
        return Machine('0/lxd/0', {'agent-state': 'unallocated',
                                   'dns-name': 'unallocated'})

    def __str__(self):
        return ("id={machine_id} state={state}, "
                "dns-name={dns_name} mem={mem} "
                "storage={storage} "
                "cpus={cpus}".format(machine_id=self.machine_id,
                                     dns_name=self.dns_name,
                                     state=self.agent_state,
                                     mem=self.mem,
                                     storage=self.storage,
                                     cpus=self.cpu_cores))

    def __repr__(self):
        return ("<Machine({dns_name},{state},{mem},"
                "{storage},{cpus})>".format(dns_name=self.dns_name,
                                            state=self.agent_state,
                                            mem=self.mem,
                                            storage=self.storage,
                                            cpus=self.cpu_cores))



# Copyright 2016 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


from urwid import (AttrMap, Divider, Pile, WidgetWrap,
                   Padding, Text, Columns)

from ubuntui.widgets.buttons import MenuSelectButton, PlainButton

from bundleplacer.assignmenttype import AssignmentType


class ServiceWidgetState(Enum):
    CHOOSING = 0
    SELECTED = 1
    UNSELECTED = 2


class SimpleServiceWidget(WidgetWrap):

    """A widget displaying a service as a button

    service - the service to display

    placement_controller - a PlacementController instance

    display_controller - a PlacerView instance

    callback - a function to be called when either of the buttons is
    pressed. The service will be passed to the function as userdata.

    show_placements - display the machine(s) currently assigned to
    host this service, both planned deployments (aka 'assignments',
    and already-deployed, called 'deployments').

    """

    def __init__(self, service, placement_controller,
                 display_controller, show_placements=False):
        self.service = service
        self.placement_controller = placement_controller
        self.display_controller = display_controller
        self.show_placements = show_placements
        self.state = ServiceWidgetState.UNSELECTED
        w = self.build_widgets()
        super().__init__(w)
        self.update()

    def selectable(self):
        return True

    def build_widgets(self):

        if self.service.subordinate:
            self.button = MenuSelectButton("I AM A SUBORDINATE SERVICE")
        else:
            self.button = MenuSelectButton("I AM A SERVICE", self.do_select)

        self.action_button_cols = Columns([], dividechars=1)
        self.action_buttons = []

        self.pile = Pile([self.button])
        return self.pile

    def get_markup(self):
        if self.service.subordinate:
            return [self.service.service_name +
                    " (subordinate)\n  " +
                    self.service.charm_source], []

        nr = self.service.required_num_units()
        pl = "s" if nr > 1 else ""
        title_markup = [self.service.service_name +
                        "\n  {} unit{}: ".format(nr, pl) +
                        self.service.charm_source]
        info_markup = []

        if not self.display_controller.has_maas:
            return title_markup, info_markup

        pd = self.placement_controller.get_assignments(self.service)
        nplaced = sum([len(pd[k]) for k in pd])

        if nr-nplaced > 0:
            pl = ""
            if nr-nplaced > 1:
                pl = "s"
            info_str = ("  {} unit{} will be auto-placed "
                        "by Juju\n".format(nr-nplaced, pl))

            info_markup.append(info_str)

        def string_for_placement_dict(d):
            if self.display_controller.has_maas:
                defstring = "Bare Metal (Default)"
            else:
                defstring = "LXD (Default)"
            s = []
            for atype, ml in sorted(d.items()):
                if atype == AssignmentType.DEFAULT:
                    aname = defstring
                else:
                    aname = atype.name

                hostnames = [m.hostname for m in ml]
                s.append("    {}: {}".format(aname,
                                             ", ".join(hostnames)))
            if len(s) == 0:
                return []
            return "\n".join(s)

        ad = self.placement_controller.get_assignments(self.service)
        info_markup += string_for_placement_dict(ad)
        return title_markup, info_markup

    def update_choosing(self):
        title_markup, _ = self.get_markup()
        msg = Padding(Text(title_markup), left=2, right=2, align='center')
        self.pile.contents = [(msg, self.pile.options()),
                              (self.action_button_cols,
                               self.pile.options()),
                              (Divider(), self.pile.options())]

    def update_default(self):
        title_markup, info_markup = self.get_markup()
        self.button.set_label(title_markup + ["\n"] + info_markup)
        if self.state == ServiceWidgetState.SELECTED:
            b = AttrMap(self.button, 'deploy_highlight_start',
                        'button_secondary focus')
        else:
            b = AttrMap(self.button, 'text', 'button_secondary focus')

        self.pile.contents = [(b, self.pile.options()),
                              (Divider(), self.pile.options())]

    def update(self):
        self.update_action_buttons()

        if self.state == ServiceWidgetState.CHOOSING:
            self.update_choosing()
        else:
            self.update_default()

    def keypress(self, size, key):
        if key == 'backspace':
            self.display_controller.remove_service(self.service)

        return super().keypress(size, key)

    def update_action_buttons(self):
        all_actions = []
        if self.display_controller.has_maas:
            all_actions = [('Choose Placement',
                            self.handle_placement_button_pressed)]
        all_actions.append(('Edit Relations',
                            self.handle_relation_button_pressed))

        self.action_buttons = [AttrMap(PlainButton(label, on_press=func),
                                       'button_secondary',
                                       'button_secondary focus')
                               for label, func in all_actions]

        self.action_buttons.append(AttrMap(
            PlainButton("Cancel",
                        on_press=self.do_cancel),
            'button_secondary',
            'button_secondary focus'))

        opts = self.action_button_cols.options()
        self.action_button_cols.contents = [(b, opts) for
                                            b in self.action_buttons]

    def do_select(self, sender):
        self.display_controller.clear_selections()
        if self.state == ServiceWidgetState.SELECTED:
            self.state = ServiceWidgetState.UNSELECTED
            self.display_controller.set_selected_service(None)
        else:
            self.display_controller.set_selected_service(self.service)
            self.state = ServiceWidgetState.CHOOSING
            self.pile.focus_position = 1
            self.action_button_cols.focus_position = 0
        self.update()

    def do_cancel(self, sender):
        self.state = ServiceWidgetState.UNSELECTED
        self.update()
        self.display_controller.show_default_view()
        self.pile.focus_position = 0

    def handle_placement_button_pressed(self, sender):
        self.state = ServiceWidgetState.SELECTED
        self.update()
        self.display_controller.edit_placement()
        self.pile.focus_position = 0

    def handle_relation_button_pressed(self, sender):
        self.state = ServiceWidgetState.SELECTED
        self.update()
        self.display_controller.edit_relations()
        self.pile.focus_position = 0



""" Service info widget, attaches properties from service to
urwid Text widgets
"""
import logging
from urwid import WidgetWrap, Text

log = logging.getLogger('serviceinfo_w')


class UnitInfoWidget(WidgetWrap):
    def __init__(self, name, unit):
        self.name = Text(name)
        self.machine = Text(unit.get('machine', ''))
        self.public_address = Text(unit.get('public-address', ''))
        self.workload_status = Text(unit.get('workload-status', ''))


class ServiceStatusWidget(WidgetWrap):
    def __init__(self, service_status):
        self.current = Text(service_status.get('current', ''))
        self.message = Text(service_status.get('message', ''))
        self.since = Text(service_status.get('since', ''))


class ServiceInfoWidget(WidgetWrap):
    def __init__(self, name, service):
        self.name = name
        for k in service.keys():
            if k == 'service-status':
                self.service_status = ServiceStatusWidget(service[k])
            elif k == 'units':
                self.units = [UnitInfoWidget(n, info) for n, info
                              in service.get('units').items()]
            else:
                setattr(self, k.replace('-', '_'), Text(service['k']))
        self.icon = Text("")



# Copyright 2014-2016 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Logging interface
"""

from __future__ import unicode_literals
import logging
import os
import pprint

from logging.handlers import TimedRotatingFileHandler


class PrettyLog():

    def __init__(self, obj):
        self.obj = obj

    def __repr__(self):
        return pprint.pformat(self.obj)


def setup_logger(name=__name__, cfg_path='.'):
    """setup logging

    """
    if not os.path.isdir(cfg_path):
        os.makedirs(cfg_path)
    LOGFILE = os.path.join(cfg_path, 'placement.log')
    commandslog = TimedRotatingFileHandler(LOGFILE,
                                           when='D',
                                           interval=1,
                                           backupCount=7)

    commandslog.setLevel('DEBUG')
    commandslog.setFormatter(logging.Formatter(
        "[%(levelname)-4s: %(asctime)s, "
        "%(filename)s:%(lineno)d] %(message)s",
        datefmt='%m-%d %H:%M:%S'))

    logger = logging.getLogger('')
    logger.setLevel('DEBUG')

    no_filter = os.environ.get('PLACEMENT_NOFILTER', None)
    if no_filter is None:
        f = logging.Filter(name='bundleplacer')
        commandslog.addFilter(f)

    logger.addHandler(commandslog)

    return logger



from ubuntui.frame import Frame  # noqa
from ubuntui.views import ErrorView
from conjure import async
from ubuntui.ev import EventLoop
import errno


class ConjureUI(Frame):
    def show_exception_message(self, ex):
        if isinstance(ex, async.ThreadCancelledException):
            pass
        elif hasattr(ex, 'errno') and ex.errno == errno.ENOENT:
            # handle oserror
            self.frame.body = ErrorView(ex.args[1])
        else:
            self.frame.body = ErrorView(ex.args[0])

        EventLoop.remove_alarms()

    def show_error_message(self, msg):
        self.frame.body = ErrorView(msg)



import time
import os.path as path
import requests
import logging
import threading
from .errors import (LoginError,
                     CharmNotFoundError,
                     RequestTimeout,
                     ServerError,
                     BadResponseError,
                     MacumbaError)
from .ws import JujuWS

log = logging.getLogger('macumba')


def query_cs(charm):
    """ This helper routine will query the charm store to pull latest revisions
    and charmstore url for the api.

    :param str charm: charm name, can be in the form of 'precise/<charm>' to
                      specify an alternate series.
    """
    try:
        series, charm = charm.split('/')
    except ValueError:
        series = 'trusty'
    charm_store_url = 'https://manage.jujucharms.com/api/3/charm'
    url = path.join(charm_store_url, series, charm)
    r = requests.get(url)
    if r.status_code != 200:
        log.error("could not find charm store URL for charm '{}'".format(url))
        rj = r.json()
        raise CharmNotFoundError("{type} {charm_id}".format(**rj))

    return r.json()


class Base:
    """ Base api class
    """
    API_VERSION = None
    FACADE_VERSIONS = {}

    def __init__(self, url, password, user='user-admin'):
        """ init

        Params:
        url: URL in form of wss://{api-endpoint}/model/{uuid}/api
        password: Password for user
        user: juju user with access to endpoint
        """
        self.url = url
        self.password = password
        self.connlock = threading.RLock()
        with self.connlock:
            self.conn = JujuWS(url, password)

        self.creds = {'Type': 'Admin',
                      'Version': 3,
                      'Request': 'Login',
                      'RequestId': 1,
                      'Params': {'auth-tag': user,
                                 'credentials': password}}

    def _prepare_strparams(self, d):
        r = {}
        for k, v in d.items():
            r[k] = str(v)
        return r

    def _prepare_constraints(self, constraints):
        for k in ['cpu-cores', 'cpu-power', 'mem', 'root-disk']:
            if constraints.get(k):
                constraints[k] = int(constraints[k])
        return constraints

    def login(self):
        """Connect and log in to juju websocket endpoint.

        block other threads until done.
        """
        with self.connlock:
            req_id = self.conn.do_connect(self.creds)
            try:
                res = self.receive(req_id)
                if 'Error' in res:
                    raise LoginError(res['ErrorCode'])
            except Exception as e:
                raise LoginError(str(e))

    def reconnect(self):
        with self.connlock:
            self.close()
            start_id = self.conn.get_current_request_id() + 1
            self.conn = JujuWS(self.url,
                               self.password,
                               start_reqid=start_id)
            self.login()

    def close(self):
        """ Closes connection to juju websocket """
        with self.connlock:
            self.conn.do_close()

    def receive(self, request_id, timeout=None):
        """receives expected message.

        returns parsed response object.

        if timeout is set, raises RequestTimeout after 'timeout' seconds
        with no received message.

        """
        res = None
        start_time = time.time()
        while res is None:
            with self.connlock:
                res = self.conn.do_receive(request_id)
            if res is None:
                time.sleep(0.1)
                if timeout and (time.time() - start_time > timeout):
                    raise RequestTimeout(request_id)

        if 'Error' in res:
            raise ServerError(res['Error'], res)

        try:
            return res['Response']
        except:
            raise BadResponseError("Failed to parse response: {}".format(res))

    def call(self, params, timeout=None):
        """ Get json data from juju api daemon.

        :params params: Additional params to be passed into request
        :type params: dict
        """
        if params['Type'] in self.FACADE_VERSIONS:
            params.update({'Version': self.FACADE_VERSIONS[params['Type']]})
        else:
            raise MacumbaError(
                'Unknown facade type: {}'.format(params['Type']))
        with self.connlock:
            req_id = self.conn.do_send(params)

        return self.receive(req_id, timeout)



""" Async Handler
Provides async operations for various api calls and other non-blocking
work.
"""

import logging
from concurrent.futures import ThreadPoolExecutor
from threading import Event
import time
log = logging.getLogger("async")


class ThreadCancelledException(Exception):
    """Exception meaning intentional cancellation"""

AsyncPool = ThreadPoolExecutor(1)
log.debug('AsyncPool={}'.format(AsyncPool))


ShutdownEvent = Event()


def submit(func, exc_callback):
    def cb(cb_f):
        e = cb_f.exception()
        if e:
            exc_callback(e)
    if ShutdownEvent.is_set():
        log.debug("ignoring async.submit due to impending shutdown.")
        return
    f = AsyncPool.submit(func)
    f.add_done_callback(cb)
    return f


def shutdown():
    ShutdownEvent.set()
    AsyncPool.shutdown(wait=False)


def sleep_until(s):
    """returns after 's' seconds.
    If the ShutdownEvent is raised before the wait is over,
    raises a ThreadCancelledException.
    """
    start = time.time()
    while not ShutdownEvent.wait(timeout=.1):
        if time.time() - start >= s:
            return True
    raise ThreadCancelledException("Thread cancelled while sleeping")
